#
# Perform a proxy request for photons
#

import sys
import logging
from sliderule import icesat2
from utils import initialize_client, display_statistics

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # Configure Logging
    logging.basicConfig(level=logging.INFO)

    # Initialize Client
    parms, cfg = initialize_client(sys.argv)

    # Request ATL06 Data
    gdf = icesat2.atl03sp(parms, asset=cfg["asset"], resources=[cfg["resource"]])

    # Display Statistics
    display_statistics(gdf, "photons")



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import netrc
import requests
import json
import struct
import ctypes
import time
import logging
import numpy
from datetime import datetime, timedelta
from sliderule import version

###############################################################################
# GLOBALS
###############################################################################

PUBLIC_URL = "slideruleearth.io"
PUBLIC_ORG = "sliderule"

service_url = PUBLIC_URL
service_org = PUBLIC_ORG

session = requests.Session()
session.trust_env = False

ps_refresh_token = None
ps_access_token = None
ps_token_exp = None

verbose = False

request_timeout = (10, 60) # (connection, read) in seconds

logger = logging.getLogger(__name__)

recdef_table = {}

arrow_file_table = {}

gps_epoch = datetime(1980, 1, 6)
tai_epoch = datetime(1970, 1, 1, 0, 0, 10)

eventformats = {
    "TEXT":     0,
    "JSON":     1
}

eventlogger = {
    0: logger.debug,
    1: logger.info,
    2: logger.warning,
    3: logger.error,
    4: logger.critical
}

datatypes = {
    "TEXT":     0,
    "REAL":     1,
    "INTEGER":  2,
    "DYNAMIC":  3
}

basictypes = {
    "INT8":     { "fmt": 'b', "size": 1, "nptype": numpy.int8   },
    "INT16":    { "fmt": 'h', "size": 2, "nptype": numpy.int16  },
    "INT32":    { "fmt": 'i', "size": 4, "nptype": numpy.int32  },
    "INT64":    { "fmt": 'q', "size": 8, "nptype": numpy.int64  },
    "UINT8":    { "fmt": 'B', "size": 1, "nptype": numpy.uint8  },
    "UINT16":   { "fmt": 'H', "size": 2, "nptype": numpy.uint16 },
    "UINT32":   { "fmt": 'I', "size": 4, "nptype": numpy.uint32 },
    "UINT64":   { "fmt": 'Q', "size": 8, "nptype": numpy.uint64 },
    "BITFIELD": { "fmt": 'x', "size": 0, "nptype": numpy.byte   },  # unsupported
    "FLOAT":    { "fmt": 'f', "size": 4, "nptype": numpy.single },
    "DOUBLE":   { "fmt": 'd', "size": 8, "nptype": numpy.double },
    "TIME8":    { "fmt": 'Q', "size": 8, "nptype": numpy.byte   },
    "STRING":   { "fmt": 's', "size": 1, "nptype": numpy.byte   }
}

codedtype2str = {
    0:  "INT8",
    1:  "INT16",
    2:  "INT32",
    3:  "INT64",
    4:  "UINT8",
    5:  "UINT16",
    6:  "UINT32",
    7:  "UINT64",
    8:  "BITFIELD",
    9:  "FLOAT",
    10: "DOUBLE",
    11: "TIME8",
    12: "STRING"
}

###############################################################################
# CLIENT EXCEPTIONS
###############################################################################

class FatalError(RuntimeError):
    pass

class TransientError(RuntimeError):
    pass

###############################################################################
# UTILITIES
###############################################################################

#
#  __populate
#
def __populate(rectype):
    global recdef_table
    if rectype not in recdef_table:
        recdef_table[rectype] = source("definition", {"rectype" : rectype})
    return recdef_table[rectype]

#
#  __parse_json
#
def __parse_json(data):
    """
    data: request response
    """
    lines = []
    for line in data.iter_content(None):
        lines.append(line)
    response = b''.join(lines)
    return json.loads(response)

#
#  __decode_native
#
def __decode_native(rectype, rawdata):
    """
    rectype: record type supplied in response (string)
    rawdata: payload supplied in response (byte array)
    """
    global recdef_table

    # initialize record
    rec = { "__rectype": rectype }

    # get/populate record definition #
    recdef = __populate(rectype)

    # iterate through each field in definition
    for fieldname in recdef.keys():

        # double underline (__) as prefix indicates meta data
        if fieldname.find("__") == 0:
            continue

        # get field properties
        field = recdef[fieldname]
        ftype = field["type"]
        offset = int(field["offset"] / 8)
        elems = field["elements"]
        flags = field["flags"]

        # do not process pointers
        if "PTR" in flags:
            continue

        # get endianess
        if "LE" in flags:
            endian = '<'
        else:
            endian = '>'

        # decode basic type
        if ftype in basictypes:

            # check if array
            is_array = not (elems == 1)

            # get number of elements
            if elems <= 0:
                elems = int((len(rawdata) - offset) / basictypes[ftype]["size"])

            # build format string
            fmt = endian + str(elems) + basictypes[ftype]["fmt"]

            # parse data
            value = struct.unpack_from(fmt, rawdata, offset)

            # set field
            if ftype == "STRING":
                rec[fieldname] = ctypes.create_string_buffer(value[0]).value.decode('ascii')
            elif is_array:
                rec[fieldname] = value
            else:
                rec[fieldname] = value[0]

        # decode user type
        else:

            # populate record definition (if needed) #
            subrecdef = __populate(ftype)

            # check if array
            is_array = not (elems == 1)

            # get number of elements
            if elems <= 0:
                elems = int((len(rawdata) - offset) / subrecdef["__datasize"])

            # return parsed data
            if is_array:
                rec[fieldname] = []
                for e in range(elems):
                    rec[fieldname].append(__decode_native(ftype, rawdata[offset:]))
                    offset += subrecdef["__datasize"]
            else:
                rec[fieldname] = __decode_native(ftype, rawdata[offset:])

    # return record #
    return rec

#
#  __parse_native
#
def __parse_native(data, callbacks):
    """
    data: request response
    """
    recs = []

    rec_hdr_size = 8
    rec_size_index = 0
    rec_size_rsps = []

    rec_size = 0
    rec_index = 0
    rec_rsps = []

    for line in data.iter_content(None):

        i = 0
        while i < len(line):

            # Parse Record Size
            if(rec_size_index < rec_hdr_size):
                bytes_available = len(line) - i
                bytes_remaining = rec_hdr_size - rec_size_index
                bytes_to_append = min(bytes_available, bytes_remaining)
                rec_size_rsps.append(line[i:i+bytes_to_append])
                rec_size_index += bytes_to_append
                if(rec_size_index >= rec_hdr_size):
                    raw = b''.join(rec_size_rsps)
                    rec_version, rec_type_size, rec_data_size = struct.unpack('>hhi', raw)
                    if rec_version != 2:
                        raise FatalError("Invalid record format: %d" % (rec_version))
                    rec_size = rec_type_size + rec_data_size
                    rec_size_rsps.clear()
                i += bytes_to_append

            # Parse Record
            elif(rec_size > 0):
                bytes_available = len(line) - i
                bytes_remaining = rec_size - rec_index
                bytes_to_append = min(bytes_available, bytes_remaining)
                rec_rsps.append(line[i:i+bytes_to_append])
                rec_index += bytes_to_append
                if(rec_index >= rec_size):
                    # Decode Record
                    rawbits = b''.join(rec_rsps)
                    rectype = ctypes.create_string_buffer(rawbits).value.decode('ascii')
                    rawdata = rawbits[len(rectype) + 1:]
                    rec     = __decode_native(rectype, rawdata)
                    if callbacks != None and rectype in callbacks:
                        # Execute Call-Back on Record
                        callbacks[rectype](rec)
                    else:
                        # Append Record
                        recs.append(rec)
                    # Reset Record Parsing
                    rec_rsps.clear()
                    rec_size_index = 0
                    rec_size = 0
                    rec_index = 0
                i += bytes_to_append

            # Zero Sized Record
            else:
                rec_size_index = 0
                rec_index = 0

    return recs

#
#  __build_auth_header
#
def __build_auth_header():
    """
    Build authentication header for use with provisioning system
    """

    global service_url, ps_access_token, ps_refresh_token, ps_token_exp
    headers = None
    if ps_access_token:
        # Check if Refresh Needed
        if time.time() > ps_token_exp:
            host = "https://ps." + service_url + "/api/org_token/refresh/"
            rqst = {"refresh": ps_refresh_token}
            hdrs = {'Content-Type': 'application/json', 'Authorization': 'Bearer ' + ps_access_token}
            rsps = session.post(host, data=json.dumps(rqst), headers=hdrs, timeout=request_timeout).json()
            ps_refresh_token = rsps["refresh"]
            ps_access_token = rsps["access"]
            ps_token_exp =  time.time() + (float(rsps["access_lifetime"]) / 2)
        # Build Authentication Header
        headers = {'Authorization': 'Bearer ' + ps_access_token}
    return headers


###############################################################################
# Default Record Processing
###############################################################################

#
#  __logeventrec
#
def __logeventrec(rec):
    if verbose:
        eventlogger[rec['level']]('%s' % (rec["attr"]))

#
#  __exceptrec
#
def __exceptrec(rec):
    if verbose:
        if rec["code"] >= 0:
            eventlogger[rec["level"]]("Exception <%d>: %s", rec["code"], rec["text"])
        else:
            eventlogger[rec["level"]]("%s", rec["text"])

#
#  _arrowrec
#
def __arrowrec(rec):
    global arrow_file_table
    try :
        filename = rec["filename"]
        if rec["__rectype"] == 'arrowrec.meta':
            if filename in arrow_file_table:
                raise FatalError("file transfer already in progress")
            arrow_file_table[filename] = { "fp": open(filename, "wb"), "size": rec["size"], "progress": 0 }
        else: # rec["__rectype"] == 'arrowrec.data'
            data = rec['data']
            file = arrow_file_table[filename]
            file["fp"].write(bytearray(data))
            file["progress"] += len(data)
            if file["progress"] >= file["size"]:
                file["fp"].close()
                del arrow_file_table[filename]
    except Exception as e:
        raise FatalError("Failed to process arrow file: {}".format(e))

#
#  Globals
#
__callbacks = {'eventrec': __logeventrec, 'exceptrec': __exceptrec, 'arrowrec.meta': __arrowrec, 'arrowrec.data': __arrowrec }

###############################################################################
# APIs
###############################################################################

#
#  source
#
def source (api, parm={}, stream=False, callbacks={}, path="/source"):
    '''
    Perform API call to SlideRule service

    Parameters
    ----------
        api:        str
                    name of the SlideRule endpoint
        parm:       dict
                    dictionary of request parameters
        stream:     bool
                    whether the request is a **normal** service or a **stream** service (see `De-serialization <./SlideRule.html#de-serialization>`_ for more details)
        callbacks:  dict
                    record type callbacks (advanced use)
        path:       str
                    path to api being requested

    Returns
    -------
    dictionary
        response data

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_url("slideruleearth.io")
        >>> rqst = {
        ...     "time": "NOW",
        ...     "input": "NOW",
        ...     "output": "GPS"
        ... }
        >>> rsps = sliderule.source("time", rqst)
        >>> print(rsps)
        {'time': 1300556199523.0, 'format': 'GPS'}
    '''
    global service_url, service_org
    rqst = json.dumps(parm)
    rsps = {}
    headers = None
    # Build Callbacks
    for c in __callbacks:
        if c not in callbacks:
            callbacks[c] = __callbacks[c]
    # Attempt Request
    try:
        # Construct Request URL and Authorization
        if service_org:
            url = 'https://%s.%s%s/%s' % (service_org, service_url, path, api)
            headers = __build_auth_header()
        else:
            url = 'http://%s%s/%s' % (service_url, path, api)
        # Perform Request
        if not stream:
            data = session.get(url, data=rqst, headers=headers, timeout=request_timeout)
        else:
            data = session.post(url, data=rqst, headers=headers, timeout=request_timeout, stream=True)
        data.raise_for_status()
        # Parse Response
        format = data.headers['Content-Type']
        if format == 'text/plain':
            rsps = __parse_json(data)
        elif format == 'application/json':
            rsps = __parse_json(data)
        elif format == 'application/octet-stream':
            rsps = __parse_native(data, callbacks)
        else:
            raise FatalError('unsupported content type: %s' % (format))
    except requests.exceptions.SSLError as e:
        raise FatalError("Unable to verify SSL certificate: {}".format(e))
    except requests.ConnectionError as e:
        raise FatalError("Connection error to endpoint {}".format(url))
    except requests.Timeout as e:
        raise TransientError("Timed-out waiting for response from endpoint {}".format(url))
    except requests.exceptions.ChunkedEncodingError as e:
        raise RuntimeError("Unexpected termination of response from endpoint {}".format(url))
    except requests.HTTPError as e:
        if e.response.status_code == 503:
            raise TransientError("Server experiencing heavy load, stalling on request to {}".format(url))
        else:
            raise FatalError("HTTP error {} from endpoint {}".format(e.response.status_code, url))
    except:
        raise
    # Return Response
    return rsps

#
#  set_url
#
def set_url (url):
    '''
    Configure sliderule package with URL of service

    Parameters
    ----------
        urls:   str
                IP address or hostname of SlideRule service (note, there is a special case where the url is provided as a list of strings
                instead of just a string; when a list is provided, the client hardcodes the set of servers that are used to process requests
                to the exact set provided; this is used for testing and for local installations and can be ignored by most users)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_url("service.my-sliderule-server.org")
    '''
    global service_url
    service_url = url

#
#  set_verbose
#
def set_verbose (enable):
    '''
    Configure sliderule package for verbose logging

    Parameters
    ----------
        enable:     bool
                    whether or not user level log messages received from SlideRule generate a Python log message

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_verbose(True)

        The default behavior of Python log messages is for them to be displayed to standard output.
        If you want more control over the behavior of the log messages being display, create and configure a Python log handler as shown below:

        >>> # import packages
        >>> import logging
        >>> from sliderule import sliderule
        >>> # Configure Logging
        >>> sliderule_logger = logging.getLogger("sliderule.sliderule")
        >>> sliderule_logger.setLevel(logging.INFO)
        >>> # Create Console Output
        >>> ch = logging.StreamHandler()
        >>> ch.setLevel(logging.INFO)
        >>> sliderule_logger.addHandler(ch)
    '''
    global verbose
    verbose = (enable == True)

#
# set_rqst_timeout
#
def set_rqst_timeout (timeout):
    '''
    Sets the TCP/IP connection and reading timeouts for future requests made to sliderule servers.
    Setting it lower means the client will failover more quickly, but may generate false positives if a processing request stalls or takes a long time returning data.
    Setting it higher means the client will wait longer before designating it a failed request which in the presence of a persistent failure means it will take longer for the client to remove the node from its available servers list.

    Parameters
    ----------
        timeout:    tuple
                    (<connection timeout in seconds>, <read timeout in seconds>)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_rqst_timeout((10, 60))
    '''
    global request_timeout
    if type(timeout) == tuple:
        request_timeout = timeout
    else:
        raise FatalError('timeout must be a tuple (<connection timeout>, <read timeout>)')

#
# update_available_servers
#
def update_available_servers (desired_nodes=None, time_to_live=None):
    '''
    Manages the number of servers in the cluster.
    If the desired_nodes parameter is set, then a request is made to change the number of servers in the cluster to the number specified.
    In all cases, the number of nodes currently running in the cluster are returned - even if desired_nodes is set;
    subsequent calls to this function is needed to check when the current number of nodes matches the desired_nodes.

    Parameters
    ----------
        desired_nodes:  int
                        the desired number of nodes in the cluster
        time_to_live:   int
                        number of minutes for the desired nodes to run

    Returns
    -------
    int
        number of nodes currently in the cluster
    int
        number of nodes available for work in the cluster

    Examples
    --------
        >>> import sliderule
        >>> num_servers, max_workers = sliderule.update_available_servers(10)
    '''

    global service_url, service_org, request_timeout

    # Update number of nodes
    if type(desired_nodes) == int:
        headers = __build_auth_header()
        if type(time_to_live) == int:
            host = "https://ps." + service_url + "/api/desired_org_num_nodes_ttl/" + service_org + "/" + str(desired_nodes) + "/" + str(time_to_live) + "/"
            rsps = session.post(host, headers=headers, timeout=request_timeout)
        else:
            host = "https://ps." + service_url + "/api/desired_org_num_nodes/" + service_org + "/" + str(desired_nodes) + "/"
            rsps = session.put(host, headers=headers, timeout=request_timeout)
        rsps.raise_for_status()

    # Get number of nodes currently registered
    rsps = source("status", parm={"service":"sliderule"}, path="/discovery")
    available_servers = rsps["nodes"]
    return available_servers, available_servers

#
# authenticate
#
def authenticate (ps_organization, ps_username=None, ps_password=None):
    '''
    Authenticate to SlideRule Provisioning System
    The username and password can be provided the following way in order of priority:
    (1) The passed in arguments `ps_username' and 'ps_password';
    (2) The O.S. environment variables 'PS_USERNAME' and 'PS_PASSWORD';
    (3) The `ps.<url>` entry in the .netrc file in your home directory

    Parameters
    ----------
        ps_organization:    str
                            name of the SlideRule organization the user belongs to

        ps_username:        str
                            SlideRule provisioning system account name

        ps_password:        str
                            SlideRule provisioning system account password
    Returns
    -------
    status
        True of successful, False if unsuccessful

    Examples
    --------
        >>> import sliderule
        >>> sliderule.authenticate("myorg")
        True
    '''
    global service_org, ps_refresh_token, ps_access_token, ps_token_exp
    login_status = False
    ps_url = "ps." + service_url

    # set organization on any authentication request
    service_org = ps_organization

    # check for direct or public access
    if service_org == None:
        return True

    # attempt retrieving from environment
    if not ps_username or not ps_password:
        ps_username = os.environ.get("PS_USERNAME")
        ps_password = os.environ.get("PS_PASSWORD")

    # attempt retrieving from netrc file
    if not ps_username or not ps_password:
        try:
            netrc_file = netrc.netrc()
            login_credentials = netrc_file.hosts[ps_url]
            ps_username = login_credentials[0]
            ps_password = login_credentials[2]
        except Exception as e:
            logger.warning("Failed to retrieve username and password from netrc file: {}".format(e))

    # authenticate to provisioning system
    if ps_username and ps_password:
        rqst = {"username": ps_username, "password": ps_password, "org_name": ps_organization}
        headers = {'Content-Type': 'application/json'}
        try:
            api = "https://" + ps_url + "/api/org_token/"
            rsps = session.post(api, data=json.dumps(rqst), headers=headers, timeout=request_timeout)
            rsps.raise_for_status()
            rsps = rsps.json()
            ps_refresh_token = rsps["refresh"]
            ps_access_token = rsps["access"]
            ps_token_exp =  time.time() + (float(rsps["access_lifetime"]) / 2)
            login_status = True
        except:
            logger.error("Unable to authenticate user %s to %s" % (ps_username, api))

    # return login status
    return login_status

#
# gps2utc
#
def gps2utc (gps_time, as_str=True, epoch=gps_epoch):
    '''
    Convert a GPS based time returned from SlideRule into a UTC time.

    Parameters
    ----------
        gps_time:   int
                    number of seconds since GPS epoch (January 6, 1980)
        as_str:     bool
                    if True, returns the time as a string; if False, returns the time as datatime object
        epoch:      datetime
                    the epoch used in the conversion, defaults to GPS epoch (Jan 6, 1980)

    Returns
    -------
    datetime
        UTC time (i.e. GMT, or Zulu time)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.gps2utc(1235331234)
        '2019-02-27 19:34:03'
    '''
    gps_time = epoch + timedelta(seconds=gps_time)
    tai_time = gps_time + timedelta(seconds=19)
    tai_timestamp = (tai_time - tai_epoch).total_seconds()
    utc_timestamp = datetime.utcfromtimestamp(tai_timestamp)
    if as_str:
        return str(utc_timestamp)
    else:
        return utc_timestamp

#
# get_definition
#
def get_definition (rectype, fieldname):
    '''
    Get the underlying format specification of a field in a return record.

    Parameters
    ----------
    rectype:    str
                the name of the type of the record (i.e. "atl03rec")
    fieldname:  str
                the name of the record field (i.e. "cycle")

    Returns
    -------
    dict
        description of each field; see the `sliderule.basictypes` variable for different field types

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_url("slideruleearth.io")
        >>> sliderule.get_definition("atl03rec", "cycle")
        {'fmt': 'H', 'size': 2, 'nptype': <class 'numpy.uint16'>}
    '''
    recdef = __populate(rectype)
    if fieldname in recdef and recdef[fieldname]["type"] in basictypes:
        return basictypes[recdef[fieldname]["type"]]
    else:
        return {}

#
# get_version
#
def get_version ():
    '''
    Get the version information for the running servers and Python client

    Returns
    -------
    dict
        dictionary of version information
    '''
    rsps = source("version", {})
    rsps["client"] = {"version": version.full_version}
    return rsps

#
# check_version
#
def check_version (plugins=[]):
    '''
    Check that the version of the client matches the version of the server and any additionally requested plugins

    Parameters
    ----------
    plugins:    list
                list of package names (as strings) to check the version on

    Returns
    -------
    bool
        True if at least minor version matches; False if major or minor version doesn't match
    '''
    status = True
    info = get_version()
    # populate version info
    versions = {}
    for entity in ['server', 'client'] + plugins:
        s = info[entity]['version'][1:].split('.')
        versions[entity] = (int(s[0]), int(s[1]), int(s[2]))
    # check major version mismatches
    if versions['server'][0] != versions['client'][0]:
        raise RuntimeError("Client (version {}) is incompatible with the server (version {})".format(versions['server'], versions['client']))
    else:
        for pkg in plugins:
            if versions[pkg][0] != versions['client'][0]:
                raise RuntimeError("Client (version {}) is incompatible with the {} plugin (version {})".format(versions['server'], pkg, versions['icesat2']))
    # check minor version mismatches
    if versions['server'][1] > versions['client'][1]:
        logger.warning("Client (version {}) is out of date with the server (version {})".format(versions['server'], versions['client']))
        status = False
    else:
        for pkg in plugins:
            if versions[pkg][1] > versions['client'][1]:
                logger.warning("Client (version {}) is out of date with the {} plugin (version {})".format(versions['server'], pkg, versions['client']))
                status = False
    # return if version check is successful
    return status



#
# Build index file (catalog) of ArcticDem hosted on AWS
import geopandas as gpd
import numpy as np
import pandas as pd
import pystac


###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # catalog_stac = 'https://pgc-opendata-dems.s3.us-west-2.amazonaws.com/pgc-data-stac.json'
    # cat = pystac.read_file(catalog_stac)
    # arcticdem_collection = cat.get_child('arcticdem')
    # mosaic_collection = arcticdem_collection.get_child('arcticdem-mosaics')

    collection_stac = 'https://pgc-opendata-dems.s3.us-west-2.amazonaws.com/arcticdem/strips/s2s041/2m.json'
    col = pystac.read_file(collection_stac)

    catalog_links = []
    collection_links = []
    item_list = []

    for link in col.links:
        if link.rel == pystac.RelType.CHILD:
            collection_links.append(link)

    print(f"Number of collection links: {len(collection_links)}")
    cnt = 0

    for link in collection_links:
        cnt += 1
        if cnt % 100 == 0:
            print(f"working collection links: {cnt}")

        if link.rel == pystac.RelType.CHILD:
            sub_cat = pystac.read_file(link.target)

            for _link in sub_cat.links:
                if _link.rel == pystac.RelType.CHILD:
                    catalog_links.append(_link)


    print(f"Number of catalog links: {len(catalog_links)}")
    cnt = 0

    for link in catalog_links:
        cnt += 1
        if cnt % 100 == 0:
            print(f"working catalog links: {cnt}")

        subcat = pystac.read_file(link)

        for _link in subcat.links:
            if _link.rel == pystac.RelType.CHILD:
                item = pystac.read_file(_link.target)
                item_list.append(item)

    print(f"Number of features: {len(item_list)}")

    # Geopandas ignores list-valued keys when opening, so this moves asset hrefs to properties for convenience
    for item in item_list:
        item.clear_links()
        asset_hrefs = pd.DataFrame(
            item.to_dict()['assets']).T['href'].to_dict()
        item.properties.update(asset_hrefs)

    items = pystac.ItemCollection(item_list)
    items.save_object('/data/ArcticDem/strips.geojson')



#
# Imports
#
import sys
from sliderule import icesat2
from utils import parse_command_line

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # Defaults
    cfg = {
        "region": "examples/grandmesa.geojson",
        "tolerance": 0.0,
        "dataset": "ATL03"
    }

    # Command line parameters
    parse_command_line(sys.argv, cfg)

    # Override region of interest
    region = icesat2.toregion(cfg["region"], cfg["tolerance"])

    # Query CMR for list of resources
    resources = icesat2.cmr(polygon=region["poly"], short_name=cfg["dataset"])
    print("Region: {} points, {} files".format(len(region["poly"]), len(resources)))
    for resource in resources:
        print(resource)



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import time
import itertools
import copy
import json
import ssl
import urllib.request
import datetime
import logging
import warnings
import numpy
import geopandas
from shapely.geometry.multipolygon import MultiPolygon
from shapely.geometry import Polygon
from sklearn.cluster import KMeans
import sliderule

###############################################################################
# GLOBALS
###############################################################################

# create logger
logger = logging.getLogger(__name__)

# profiling times for each major function
profiles = {}

# default asset
DEFAULT_ASSET="nsidc-s3"

# default standard data product version
DEFAULT_ICESAT2_SDP_VERSION='005'

# default maximum number of resources to process in one request
DEFAULT_MAX_REQUESTED_RESOURCES = 300
max_requested_resources = DEFAULT_MAX_REQUESTED_RESOURCES

# default maximum number of workers used for one request
DEFAULT_MAX_WORKERS_PER_NODE = 3

# icesat2 parameters
CNF_POSSIBLE_TEP = -2
CNF_NOT_CONSIDERED = -1
CNF_BACKGROUND = 0
CNF_WITHIN_10M = 1
CNF_SURFACE_LOW = 2
CNF_SURFACE_MEDIUM = 3
CNF_SURFACE_HIGH = 4
SRT_LAND = 0
SRT_OCEAN = 1
SRT_SEA_ICE = 2
SRT_LAND_ICE = 3
SRT_INLAND_WATER = 4
ALL_ROWS = -1
MAX_COORDS_IN_POLYGON = 16384
GT1L = 10
GT1R = 20
GT2L = 30
GT2R = 40
GT3L = 50
GT3R = 60
STRONG_SPOTS = (1, 3, 5)
WEAK_SPOTS = (2, 4, 6)
LEFT_PAIR = 0
RIGHT_PAIR = 1
SC_BACKWARD = 0
SC_FORWARD = 1

# gps-based epoch for delta times
ATLAS_SDP_EPOCH = datetime.datetime(2018, 1, 1)

###############################################################################
# NSIDC UTILITIES
###############################################################################
# The functions below have been adapted from the NSIDC download script and
# carry the following notice:
#
# Copyright (c) 2020 Regents of the University of Colorado
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.

# WGS84 / Mercator, Earth as Geoid, Coordinate system on the surface of a sphere or ellipsoid of reference.
EPSG_MERCATOR = "EPSG:4326"

CMR_URL = 'https://cmr.earthdata.nasa.gov'
CMR_PAGE_SIZE = 2000
CMR_FILE_URL = ('{0}/search/granules.json?provider=NSIDC_ECS'
                '&sort_key[]=start_date&sort_key[]=producer_granule_id'
                '&scroll=true&page_size={1}'.format(CMR_URL, CMR_PAGE_SIZE))

def __build_version_query_params(version):
    desired_pad_length = 3
    if len(version) > desired_pad_length:
        raise RuntimeError('Version string too long: "{0}"'.format(version))

    version = str(int(version))  # Strip off any leading zeros
    query_params = ''

    while len(version) <= desired_pad_length:
        padded_version = version.zfill(desired_pad_length)
        query_params += '&version={0}'.format(padded_version)
        desired_pad_length -= 1
    return query_params

def __cmr_filter_urls(search_results):
    """Select only the desired data files from CMR response."""
    if 'feed' not in search_results or 'entry' not in search_results['feed']:
        return []

    entries = [e['links']
               for e in search_results['feed']['entry']
               if 'links' in e]
    # Flatten "entries" to a simple list of links
    links = list(itertools.chain(*entries))

    urls = []
    unique_filenames = set()
    for link in links:
        if 'href' not in link:
            # Exclude links with nothing to download
            continue
        if 'inherited' in link and link['inherited'] is True:
            # Why are we excluding these links?
            continue
        if 'rel' in link and 'data#' not in link['rel']:
            # Exclude links which are not classified by CMR as "data" or "metadata"
            continue

        if 'title' in link and 'opendap' in link['title'].lower():
            # Exclude OPeNDAP links--they are responsible for many duplicates
            # This is a hack; when the metadata is updated to properly identify
            # non-datapool links, we should be able to do this in a non-hack way
            continue

        filename = link['href'].split('/')[-1]
        if filename in unique_filenames:
            # Exclude links with duplicate filenames (they would overwrite)
            continue

        unique_filenames.add(filename)

        if ".h5" in link['href'][-3:]:
            resource = link['href'].split("/")[-1]
            urls.append(resource)

    return urls

def __cmr_granule_metadata(search_results):
    """Get the metadata for CMR returned granules"""
    # GeoDataFrame with granule metadata
    granule_metadata = __emptyframe()
    # return empty dataframe if no CMR entries
    if 'feed' not in search_results or 'entry' not in search_results['feed']:
        return granule_metadata
    # for each CMR entry
    for e in search_results['feed']['entry']:
        # columns for dataframe
        columns = {}
        # time start and time end of granule
        columns['time_start'] = numpy.datetime64(e['time_start'])
        columns['time_end'] = numpy.datetime64(e['time_end'])
        columns['time_updated'] = numpy.datetime64(e['updated'])
        # get the granule size and convert to bits
        columns['granule_size'] = float(e['granule_size'])*(2.0**20)
        # Create Pandas DataFrame object
        # use granule id as index
        df = geopandas.pd.DataFrame(columns, index=[e['id']])
        # Generate Geometry Column
        if 'polygons' in e:
            coords = [float(i) for i in e['polygons'][0][0].split()]
            geometry = Polygon(zip(coords[1::2], coords[::2]))
        else:
            geometry, = geopandas.points_from_xy([None], [None])
        # Build GeoDataFrame (default geometry is crs=EPSG_MERCATOR)
        gdf = geopandas.GeoDataFrame(df, geometry=[geometry], crs=EPSG_MERCATOR)
        # append to combined GeoDataFrame and catch warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            granule_metadata = granule_metadata.append(gdf)
    # return granule metadata
    # - time start and time end
    # - time granule was updated
    # - granule size in bits
    # - polygons as geodataframe geometry
    return granule_metadata

def __cmr_search(short_name, version, time_start, time_end, **kwargs):
    """Perform a scrolling CMR query for files matching input criteria."""
    kwargs.setdefault('polygon',None)
    kwargs.setdefault('name_filter',None)
    kwargs.setdefault('return_metadata',False)
    # build params
    params = '&short_name={0}'.format(short_name)
    params += __build_version_query_params(version)
    params += '&temporal[]={0},{1}'.format(time_start, time_end)
    if kwargs['polygon']:
        params += '&polygon={0}'.format(kwargs['polygon'])
    if kwargs['name_filter']:
        params += '&options[producer_granule_id][pattern]=true'
        params += '&producer_granule_id[]=' + kwargs['name_filter']
    cmr_query_url = CMR_FILE_URL + params
    logger.debug('cmr request={0}\n'.format(cmr_query_url))

    cmr_scroll_id = None
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE

    urls = []
    # GeoDataFrame with granule metadata
    metadata = __emptyframe()
    while True:
        req = urllib.request.Request(cmr_query_url)
        if cmr_scroll_id:
            req.add_header('cmr-scroll-id', cmr_scroll_id)
        response = urllib.request.urlopen(req, context=ctx)
        if not cmr_scroll_id:
            # Python 2 and 3 have different case for the http headers
            headers = {k.lower(): v for k, v in dict(response.info()).items()}
            cmr_scroll_id = headers['cmr-scroll-id']
            hits = int(headers['cmr-hits'])
        search_page = response.read()
        search_page = json.loads(search_page.decode('utf-8'))
        url_scroll_results = __cmr_filter_urls(search_page)
        if not url_scroll_results:
            break
        urls += url_scroll_results
        # query for granule metadata and polygons
        if kwargs['return_metadata']:
            metadata_results = __cmr_granule_metadata(search_page)
        else:
            metadata_results = [None for _ in url_scroll_results]
        # append granule metadata and catch warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            metadata = metadata.append(metadata_results)

    return (urls,metadata)

###############################################################################
# LOCAL FUNCTIONS
###############################################################################

#
# Calculate Laser Spot
#
def __calcspot(sc_orient, track, pair):

    # spacecraft in forward orientation
    if sc_orient == SC_BACKWARD:
        if track == 1:
            if pair == LEFT_PAIR:
                return 1
            elif pair == RIGHT_PAIR:
                return 2
        elif track == 2:
            if pair == LEFT_PAIR:
                return 3
            elif pair == RIGHT_PAIR:
                return 4
        elif track == 3:
            if pair == LEFT_PAIR:
                return 5
            elif pair == RIGHT_PAIR:
                return 6

    # spacecraft in backward orientation
    elif sc_orient == SC_FORWARD:
        if track == 1:
            if pair == LEFT_PAIR:
                return 6
            elif pair == RIGHT_PAIR:
                return 5
        elif track == 2:
            if pair == LEFT_PAIR:
                return 4
            elif pair == RIGHT_PAIR:
                return 3
        elif track == 3:
            if pair == LEFT_PAIR:
                return 2
            elif pair == RIGHT_PAIR:
                return 1

    # unknown spot
    return 0

#
#  Get Values from Raw Buffer
#
def __get_values(data, dtype, size):
    """
    data:   tuple of bytes
    dtype:  element of codedtype
    size:   bytes in data
    """

    raw = bytes(data)
    datatype = sliderule.basictypes[sliderule.codedtype2str[dtype]]["nptype"]
    num_elements = int(size / numpy.dtype(datatype).itemsize)
    slicesize = num_elements * numpy.dtype(datatype).itemsize # truncates partial bytes
    values = numpy.frombuffer(raw[:slicesize], dtype=datatype, count=num_elements)

    return values

#
#  Query Resources from CMR
#
def __query_resources(parm, version, **kwargs):

    # Latch Start Time
    tstart = time.perf_counter()

    # Check Parameters are Valid
    if ("poly" not in parm) and ("t0" not in parm) and ("t1" not in parm):
        logger.error("Must supply some bounding parameters with request (poly, t0, t1)")
        return []

    # Submission Arguments for CMR
    kwargs['version'] = version
    kwargs.setdefault('return_metadata', False)

    # Pull Out Polygon
    if "clusters" in parm and parm["clusters"] and len(parm["clusters"]) > 0:
        kwargs['polygon'] = parm["clusters"]
    elif "poly" in parm and parm["poly"] and len(parm["poly"]) > 0:
        kwargs['polygon'] = parm["poly"]

    # Pull Out Time Period
    if "t0" in parm:
        kwargs['time_start'] = parm["t0"]
    if "t1" in parm:
        kwargs['time_end'] = parm["t1"]

    # Build Filters
    name_filter_enabled = False
    rgt_filter = '????'
    if "rgt" in parm:
        rgt_filter = f'{parm["rgt"]}'.zfill(4)
        name_filter_enabled = True
    cycle_filter = '??'
    if "cycle" in parm:
        cycle_filter = f'{parm["cycle"]}'.zfill(2)
        name_filter_enabled = True
    region_filter = '??'
    if "region" in parm:
        region_filter = f'{parm["region"]}'.zfill(2)
        name_filter_enabled = True
    if name_filter_enabled:
        kwargs['name_filter'] = '*_' + rgt_filter + cycle_filter + region_filter + '_*'

    # Make CMR Request
    if kwargs['return_metadata']:
        resources,metadata = cmr(**kwargs)
    else:
        resources = cmr(**kwargs)

    # Check Resources are Under Limit
    if(len(resources) > max_requested_resources):
        raise RuntimeError('Exceeded maximum requested granules: {} (current max is {})\nConsider using icesat2.set_max_resources to set a higher limit.'.format(len(resources), max_requested_resources))
    else:
        logger.info("Identified %d resources to process", len(resources))

    # Update Profile
    profiles[__query_resources.__name__] = time.perf_counter() - tstart

    # Return Resources
    if kwargs['return_metadata']:
        return (resources,metadata)
    else:
        return resources

#
#  Create Empty GeoDataFrame
#
def __emptyframe(**kwargs):
    # set default keyword arguments
    kwargs['crs'] = EPSG_MERCATOR
    return geopandas.GeoDataFrame(geometry=geopandas.points_from_xy([], []), crs=kwargs['crs'])

#
#  Dictionary to GeoDataFrame
#
def __todataframe(columns, delta_time_key="delta_time", lon_key="lon", lat_key="lat", **kwargs):

    # Latch Start Time
    tstart = time.perf_counter()

    # Set Default Keyword Arguments
    kwargs['index_key'] = "time"
    kwargs['crs'] = EPSG_MERCATOR

    # Check Empty Columns
    if len(columns) <= 0:
        return __emptyframe(**kwargs)

    # Generate Time Column
    delta_time = (columns[delta_time_key]*1e9).astype('timedelta64[ns]')
    atlas_sdp_epoch = numpy.datetime64(ATLAS_SDP_EPOCH)
    columns['time'] = geopandas.pd.to_datetime(atlas_sdp_epoch + delta_time)

    # Generate Geometry Column
    geometry = geopandas.points_from_xy(columns[lon_key], columns[lat_key])
    del columns[lon_key]
    del columns[lat_key]

    # Create Pandas DataFrame object
    if type(columns) == dict:
        df = geopandas.pd.DataFrame(columns)
    else:
        df = columns

    # Build GeoDataFrame (default geometry is crs=EPSG_MERCATOR)
    gdf = geopandas.GeoDataFrame(df, geometry=geometry, crs=kwargs['crs'])

    # Set index (default is Timestamp), can add `verify_integrity=True` to check for duplicates
    # Can do this during DataFrame creation, but this allows input argument for desired column
    gdf.set_index(kwargs['index_key'], inplace=True)

    # Sort values for reproducible output despite async processing
    gdf.sort_index(inplace=True)

    # Update Profile
    profiles[__todataframe.__name__] = time.perf_counter() - tstart

    # Return GeoDataFrame
    return gdf

#
# GeoDataFrame to Polygon
#
def __gdf2poly(gdf):

    # latch start time
    tstart = time.perf_counter()

    # pull out coordinates
    hull = gdf.unary_union.convex_hull
    polygon = [{"lon": coord[0], "lat": coord[1]} for coord in list(hull.exterior.coords)]

    # determine winding of polygon #
    #              (x2               -    x1)             *    (y2               +    y1)
    wind = sum([(polygon[i+1]["lon"] - polygon[i]["lon"]) * (polygon[i+1]["lat"] + polygon[i]["lat"]) for i in range(len(polygon) - 1)])
    if wind > 0:
        # reverse direction (make counter-clockwise) #
        ccw_poly = []
        for i in range(len(polygon), 0, -1):
            ccw_poly.append(polygon[i - 1])
        # replace region with counter-clockwise version #
        polygon = ccw_poly

    # Update Profile
    profiles[__gdf2poly.__name__] = time.perf_counter() - tstart

    # return polygon
    return polygon

#
# Process Output File
#
def __procoutputfile(parm, lon_key, lat_key):
    if "open_on_complete" in parm["output"] and parm["output"]["open_on_complete"]:
        # Read Parquet File as DataFrame
        df = geopandas.pd.read_parquet(parm["output"]["path"])
        # Build GeoDataFrame
        gdf = __todataframe(df, lon_key=lon_key, lat_key=lat_key)
        # Return Results
        return gdf
    else:
        # Return Parquet Filename
        return parm["output"]["path"]

###############################################################################
# APIs
###############################################################################

#
#  Initialize
#
def init (url, verbose=False, max_resources=DEFAULT_MAX_REQUESTED_RESOURCES, loglevel=logging.CRITICAL, organization=sliderule.service_org):
    '''
    Initializes the Python client for use with SlideRule, and should be called before other ICESat-2 API calls.
    This function is a wrapper for a handful of sliderule functions that would otherwise all have to be called in order to initialize the client.

    Parameters
    ----------
        url :           str
                        the IP address or hostname of the SlideRule service (note, there is a special case where the url is provided as a list of strings instead of just a string; when a list is provided, the client hardcodes the set of servers that are used to process requests to the exact set provided; this is used for testing and for local installations and can be ignored by most users)
        verbose :       bool
                        whether or not user level log messages received from SlideRule generate a Python log message (see `sliderule.set_verbose <../user_guide/SlideRule.html#set_verbose>`_)
        max_resources : int
                        the maximum number of resources that are allowed to be processed in a single request
        loglevel :      int
                        minimum severity of log message to output
        organization:   str
                        SlideRule provisioning system organization the user belongs to (see sliderule.authenticate for details)

    Examples
    --------
        >>> from sliderule import icesat2
        >>> icesat2.init("my-sliderule-service.my-company.com", True)
    '''
    set_max_resources(max_resources)
    if verbose:
        loglevel = logging.INFO
    logging.basicConfig(level=loglevel)
    sliderule.set_url(url)
    sliderule.set_verbose(verbose)
    sliderule.authenticate(organization)
    sliderule.check_version(plugins=['icesat2'])

#
#  Set Maximum Resources
#
def set_max_resources (max_resources):
    '''
    Sets the maximum allowed number of resources to be processed in one request.  This is mainly provided as a sanity check for the user.

    Parameters
    ----------
        max_resources : int
                        the maximum number of resources that are allowed to be processed in a single request

    Examples
    --------
        >>> from sliderule import icesat2
        >>> icesat2.set_max_resources(1000)
    '''
    global max_requested_resources
    max_requested_resources = max_resources

#
#  Common Metadata Repository
#
def cmr(**kwargs):
    '''
    Query the `NASA Common Metadata Repository (CMR) <https://cmr.earthdata.nasa.gov/search>`_ for a list of data within temporal and spatial parameters

    Parameters
    ----------
        polygon:    list
                    either a single list of longitude,latitude in counter-clockwise order with first and last point matching, defining region of interest (see `polygons <../user_guide/ICESat-2.html#polygons>`_), or a list of such lists when the region includes more than one polygon
        time_start: str
                    starting time for query in format ``<year>-<month>-<day>T<hour>:<minute>:<second>Z``
        time_end:   str
                    ending time for query in format ``<year>-<month>-<day>T<hour>:<minute>:<second>Z``
        version:    str
                    dataset version as found in the `NASA CMR Directory <https://cmr.earthdata.nasa.gov/search/site/collections/directory/eosdis>`_
        short_name: str
                    dataset short name as defined in the `NASA CMR Directory <https://cmr.earthdata.nasa.gov/search/site/collections/directory/eosdis>`_

    Returns
    -------
    list
        files (granules) for the dataset fitting the spatial and temporal parameters

    Examples
    --------
        >>> from sliderule import icesat2
        >>> region = [ {"lon": -108.3435200747503, "lat": 38.89102961045247},
        ...            {"lon": -107.7677425431139, "lat": 38.90611184543033},
        ...            {"lon": -107.7818591266989, "lat": 39.26613714985466},
        ...            {"lon": -108.3605610678553, "lat": 39.25086131372244},
        ...            {"lon": -108.3435200747503, "lat": 38.89102961045247} ]
        >>> granules = icesat2.cmr(polygon=region)
        >>> granules
        ['ATL03_20181017222812_02950102_003_01.h5', 'ATL03_20181110092841_06530106_003_01.h5', ... 'ATL03_20201111102237_07370902_003_01.h5']
    '''
    # set default polygon
    kwargs.setdefault('polygon', None)
    # set default start time to start of ICESat-2 mission
    kwargs.setdefault('time_start', '2018-10-13T00:00:00Z')
    # set default stop time to current time
    kwargs.setdefault('time_end', datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))
    # set default version and product short name
    kwargs.setdefault('version', DEFAULT_ICESAT2_SDP_VERSION)
    kwargs.setdefault('short_name','ATL03')
    # return metadata for each requested granule
    kwargs.setdefault('return_metadata', False)
    # set default name filter
    kwargs.setdefault('name_filter', None)

    # return value
    resources = {} # [<url>] = <polygon>

    # create list of polygons
    polygons = [None]
    if kwargs['polygon'] and len(kwargs['polygon']) > 0:
        if type(kwargs['polygon'][0]) == dict:
            polygons = [copy.deepcopy(kwargs['polygon'])]
        elif type(kwargs['polygon'][0] == list):
            polygons = copy.deepcopy(kwargs['polygon'])

    # iterate through each polygon (or none if none supplied)
    for polygon in polygons:
        urls = []
        metadata = __emptyframe()

        # issue CMR request
        for tolerance in [0.0001, 0.001, 0.01, 0.1, 1.0, None]:

            # convert polygon list into string
            polystr = None
            if polygon:
                flatpoly = []
                for p in polygon:
                    flatpoly.append(p["lon"])
                    flatpoly.append(p["lat"])
                polystr = str(flatpoly)[1:-1]
                polystr = polystr.replace(" ", "") # remove all spaces as this will be embedded in a url

            # call into NSIDC routines to make CMR request
            try:
                urls,metadata = __cmr_search(kwargs['short_name'],
                                            kwargs['version'],
                                            kwargs['time_start'],
                                            kwargs['time_end'],
                                            polygon=polystr,
                                            return_metadata=kwargs['return_metadata'],
                                            name_filter=kwargs['name_filter'])
                break # exit loop because cmr search was successful
            except urllib.error.HTTPError as e:
                logger.error('HTTP Request Error: {}'.format(e.reason))
            except RuntimeError as e:
                logger.error("Runtime Error:", e)

            # simplify polygon
            if polygon and tolerance:
                raw_multi_polygon = [[(tuple([(c['lon'], c['lat']) for c in polygon]), [])]]
                shape = MultiPolygon(*raw_multi_polygon)
                buffered_shape = shape.buffer(tolerance)
                simplified_shape = buffered_shape.simplify(tolerance)
                simplified_coords = list(simplified_shape.exterior.coords)
                logger.warning('Using simplified polygon (for CMR request only!), {} points using tolerance of {}'.format(len(simplified_coords), tolerance))
                region = []
                for coord in simplified_coords:
                    point = {"lon": coord[0], "lat": coord[1]}
                    region.insert(0,point)
                polygon = region
            else:
                break # exit here because nothing can be done

        # populate resources
        for i,url, in enumerate(urls):
            resources[url] = metadata.iloc[i]

    # build return lists
    url_list = list(resources.keys())
    meta_list = list(resources.values())

    if kwargs['return_metadata']:
        return (url_list,meta_list)
    else:
        return url_list

#
#  ATL06
#
def atl06 (parm, resource, asset=DEFAULT_ASSET):
    '''
    Performs ATL06-SR processing on ATL03 data and returns gridded elevations

    Parameters
    ----------
    parms:      dict
                parameters used to configure ATL06-SR algorithm processing (see `Parameters <../user_guide/ICESat-2.html#parameters>`_)
    resource:   str
                ATL03 HDF5 filename
    asset:      str
                data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)

    Returns
    -------
    GeoDataFrame
        gridded elevations (see `Elevations <../user_guide/ICESat-2.html#elevations>`_)
    '''
    return atl06p(parm, asset=asset, resources=[resource])

#
#  Parallel ATL06
#
def atl06p(parm, asset=DEFAULT_ASSET, version=DEFAULT_ICESAT2_SDP_VERSION, callbacks={}, resources=None):
    '''
    Performs ATL06-SR processing in parallel on ATL03 data and returns gridded elevations.  This function expects that the **parm** argument
    includes a polygon which is used to fetch all available resources from the CMR system automatically.  If **resources** is specified
    then any polygon or resource filtering options supplied in **parm** are ignored.

    Warnings
    --------
        It is often the case that the list of resources (i.e. granules) returned by the CMR system includes granules that come close, but
        do not actually intersect the region of interest.  This is due to geolocation margin added to all CMR ICESat-2 resources in order to account
        for the spacecraft off-pointing.  The consequence is that SlideRule will return no data for some of the resources and issue a warning statement
        to that effect; this can be ignored and indicates no issue with the data processing.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure ATL06-SR algorithm processing (see `Parameters <../user_guide/ICESat-2.html#parameters>`_)
        asset:          str
                        data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)
        version:        str
                        the version of the ATL03 data to use for processing
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["ATL03_20181019065445_03150111_004_01.h5", ...])

    Returns
    -------
    GeoDataFrame
        gridded elevations (see `Elevations <../user_guide/ICESat-2.html#elevations>`_)

    Examples
    --------
        >>> from sliderule import icesat2
        >>> icesat2.init("slideruleearth.io", True)
        >>> parms = { "cnf": 4, "ats": 20.0, "cnt": 10, "len": 40.0, "res": 20.0, "maxi": 1 }
        >>> resources = ["ATL03_20181019065445_03150111_003_01.h5"]
        >>> atl03_asset = "atlas-local"
        >>> rsps = icesat2.atl06p(parms, asset=atl03_asset, resources=resources)
        >>> rsps
                dh_fit_dx  w_surface_window_final  ...                       time                     geometry
        0        0.000042               61.157661  ... 2018-10-19 06:54:46.104937  POINT (-63.82088 -79.00266)
        1        0.002019               61.157683  ... 2018-10-19 06:54:46.467038  POINT (-63.82591 -79.00247)
        2        0.001783               61.157678  ... 2018-10-19 06:54:46.107756  POINT (-63.82106 -79.00283)
        3        0.000969               61.157666  ... 2018-10-19 06:54:46.469867  POINT (-63.82610 -79.00264)
        4       -0.000801               61.157665  ... 2018-10-19 06:54:46.110574  POINT (-63.82124 -79.00301)
        ...           ...                     ...  ...                        ...                          ...
        622407  -0.000970               61.157666  ... 2018-10-19 07:00:29.606632  POINT (135.57522 -78.98983)
        622408   0.004620               61.157775  ... 2018-10-19 07:00:29.250312  POINT (135.57052 -78.98983)
        622409  -0.001366               61.157671  ... 2018-10-19 07:00:29.609435  POINT (135.57504 -78.98966)
        622410  -0.004041               61.157748  ... 2018-10-19 07:00:29.253123  POINT (135.57034 -78.98966)
        622411  -0.000482               61.157663  ... 2018-10-19 07:00:29.612238  POINT (135.57485 -78.98948)

        [622412 rows x 16 columns]
    '''
    try:
        tstart = time.perf_counter()

        # Get List of Resources from CMR (if not supplied)
        if resources == None:
            resources = __query_resources(parm, version)

        # Build ATL06 Request
        rqst = {
            "atl03-asset" : asset,
            "resources": resources,
            "parms": parm
        }

        # Make API Processing Request
        rsps = sliderule.source("atl06p", rqst, stream=True, callbacks=callbacks)

        # Check for Output Options
        if "output" in parm:
            profiles[atl06p.__name__] = time.perf_counter() - tstart
            return __procoutputfile(parm, "lon", "lat")
        else: # Native Output
            # Flatten Responses
            tstart_flatten = time.perf_counter()
            columns = {}
            elevation_records = []
            num_elevations = 0
            field_dictionary = {} # ['field_name'] = {"extent_id": [], field_name: []}
            if len(rsps) > 0:
                # Sort Records
                for rsp in rsps:
                    if 'atl06rec' in rsp['__rectype']:
                        elevation_records += rsp,
                        num_elevations += len(rsp['elevation'])
                    elif 'extrec' == rsp['__rectype']:
                        field_name = parm['atl03_geo_fields'][rsp['field_index']]
                        if field_name not in field_dictionary:
                            field_dictionary[field_name] = {"extent_id": [], field_name: []}
                        # Parse Ancillary Data
                        data = __get_values(rsp['data'], rsp['data_type'], len(rsp['data']))
                        # Add Left Pair Track Entry
                        field_dictionary[field_name]['extent_id'] += rsp['extent_id'] | 0x2,
                        field_dictionary[field_name][field_name] += data[LEFT_PAIR],
                        # Add Right Pair Track Entry
                        field_dictionary[field_name]['extent_id'] += rsp['extent_id'] | 0x3,
                        field_dictionary[field_name][field_name] += data[RIGHT_PAIR],
                # Build Elevation Columns
                if num_elevations > 0:
                    # Initialize Columns
                    sample_elevation_record = elevation_records[0]["elevation"][0]
                    for field in sample_elevation_record.keys():
                        fielddef = sliderule.get_definition(sample_elevation_record['__rectype'], field)
                        if len(fielddef) > 0:
                            columns[field] = numpy.empty(num_elevations, fielddef["nptype"])
                    # Populate Columns
                    elev_cnt = 0
                    for record in elevation_records:
                        for elevation in record["elevation"]:
                            for field in columns:
                                columns[field][elev_cnt] = elevation[field]
                            elev_cnt += 1
            else:
                logger.debug("No response returned")

            profiles["flatten"] = time.perf_counter() - tstart_flatten

            # Build GeoDataFrame
            gdf = __todataframe(columns)

            # Merge Ancillary Fields
            tstart_merge = time.perf_counter()
            for field in field_dictionary:
                df = geopandas.pd.DataFrame(field_dictionary[field])
                gdf = gdf.merge(df, on='extent_id', how='inner').set_axis(gdf.index)
            profiles["merge"] = time.perf_counter() - tstart_merge

            # Delete Extent ID Column
            if len(gdf) > 0:
                del gdf["extent_id"]

            # Return Response
            profiles[atl06p.__name__] = time.perf_counter() - tstart
            return gdf

    # Handle Runtime Errors
    except RuntimeError as e:
        logger.critical(e)
        return __emptyframe()

#
#  Subsetted ATL03
#
def atl03s (parm, resource, asset=DEFAULT_ASSET):
    '''
    Subsets ATL03 data given the polygon and time range provided and returns segments of photons

    Parameters
    ----------
        parms:      dict
                    parameters used to configure ATL03 subsetting (see `Parameters <../user_guide/ICESat-2.html#parameters>`_)
        resource:   str
                    ATL03 HDF5 filename
        asset:      str
                    data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)

    Returns
    -------
    GeoDataFrame
        ATL03 extents (see `Photon Segments <../user_guide/ICESat-2.html#photon-segments>`_)
    '''
    return atl03sp(parm, asset=asset, resources=[resource])

#
#  Parallel Subsetted ATL03
#
def atl03sp(parm, asset=DEFAULT_ASSET, version=DEFAULT_ICESAT2_SDP_VERSION, callbacks={}, resources=None):
    '''
    Performs ATL03 subsetting in parallel on ATL03 data and returns photon segment data.  Unlike the `atl03s <#atl03s>`_ function,
    this function does not take a resource as a parameter; instead it is expected that the **parm** argument includes a polygon which
    is used to fetch all available resources from the CMR system automatically.

    Warnings
    --------
        Note, it is often the case that the list of resources (i.e. granules) returned by the CMR system includes granules that come close, but
        do not actually intersect the region of interest.  This is due to geolocation margin added to all CMR ICESat-2 resources in order to account
        for the spacecraft off-pointing.  The consequence is that SlideRule will return no data for some of the resources and issue a warning statement to that effect; this can be ignored and indicates no issue with the data processing.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure ATL03 subsetting (see `Parameters <../user_guide/ICESat-2.html#parameters>`_)
        asset:          str
                        data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)
        version:        str
                        the version of the ATL03 data to return
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["ATL03_20181019065445_03150111_004_01.h5", ...])

    Returns
    -------
    GeoDataFrame
        ATL03 segments (see `Photon Segments <../user_guide/ICESat-2.html#photon-segments>`_)
    '''
    try:
        tstart = time.perf_counter()

        # Get List of Resources from CMR (if not specified)
        if resources == None:
            resources = __query_resources(parm, version)

        # Build ATL03 Subsetting Request
        rqst = {
            "atl03-asset" : asset,
            "resources": resources,
            "parms": parm
        }

        # Make API Processing Request
        rsps = sliderule.source("atl03sp", rqst, stream=True, callbacks=callbacks)

        # Check for Output Options
        if "output" in parm:
            profiles[atl03sp.__name__] = time.perf_counter() - tstart
            return __procoutputfile(parm, "longitude", "latitude")
        else: # Native Output
            # Flatten Responses
            tstart_flatten = time.perf_counter()
            columns = {}
            sample_photon_record = None
            photon_records = []
            num_photons = 0
            extent_dictionary = {}
            extent_field_types = {} # ['field_name'] = nptype
            photon_dictionary = {}
            photon_field_types = {} # ['field_name'] = nptype
            if len(rsps) > 0:
                # Sort Records
                for rsp in rsps:
                    extent_id = rsp['extent_id']
                    if 'atl03rec' in rsp['__rectype']:
                        photon_records += rsp,
                        num_photons += len(rsp['data'])
                        if sample_photon_record == None and len(rsp['data']) > 0:
                            sample_photon_record = rsp
                    elif 'extrec' == rsp['__rectype']:
                        # Get Field Type
                        field_name = parm['atl03_geo_fields'][rsp['field_index']]
                        if field_name not in extent_field_types:
                            extent_field_types[field_name] = sliderule.basictypes[sliderule.codedtype2str[rsp['data_type']]]["nptype"]
                        # Initialize Extent Dictionary Entry
                        if extent_id not in extent_dictionary:
                            extent_dictionary[extent_id] = {}
                        # Save of Values per Extent ID per Field Name
                        data = __get_values(rsp['data'], rsp['data_type'], len(rsp['data']))
                        extent_dictionary[extent_id][field_name] = data
                    elif 'phrec' == rsp['__rectype']:
                        # Get Field Type
                        field_name = parm['atl03_ph_fields'][rsp['field_index']]
                        if field_name not in photon_field_types:
                            photon_field_types[field_name] = sliderule.basictypes[sliderule.codedtype2str[rsp['data_type']]]["nptype"]
                        # Initialize Extent Dictionary Entry
                        if extent_id not in photon_dictionary:
                            photon_dictionary[extent_id] = {}
                        # Save of Values per Extent ID per Field Name
                        data = __get_values(rsp['data'], rsp['data_type'], len(rsp['data']))
                        photon_dictionary[extent_id][field_name] = data
                # Build Elevation Columns
                if num_photons > 0:
                    # Initialize Columns
                    for field in sample_photon_record.keys():
                        fielddef = sliderule.get_definition("atl03rec", field)
                        if len(fielddef) > 0:
                            columns[field] = numpy.empty(num_photons, fielddef["nptype"])
                    for field in sample_photon_record["data"][0].keys():
                        fielddef = sliderule.get_definition("atl03rec.photons", field)
                        if len(fielddef) > 0:
                            columns[field] = numpy.empty(num_photons, fielddef["nptype"])
                    for field in extent_field_types.keys():
                        columns[field] = numpy.empty(num_photons, extent_field_types[field])
                    for field in photon_field_types.keys():
                        columns[field] = numpy.empty(num_photons, photon_field_types[field])
                    # Populate Columns
                    ph_cnt = 0
                    for record in photon_records:
                        ph_index = 0
                        pair = 0
                        left_cnt = record["count"][0]
                        extent_id = record['extent_id']
                        # Get Extent Fields to Add to Extent
                        extent_field_dictionary = {}
                        if extent_id in extent_dictionary:
                            extent_field_dictionary = extent_dictionary[extent_id]
                        # Get Photon Fields to Add to Extent
                        photon_field_dictionary = {}
                        if extent_id in photon_dictionary:
                            photon_field_dictionary = photon_dictionary[extent_id]
                        # For Each Photon in Extent
                        for photon in record["data"]:
                            if ph_index >= left_cnt:
                                pair = 1
                            # Add per Extent Fields
                            for field in record.keys():
                                if field in columns:
                                    if field == "count":
                                        columns[field][ph_cnt] = pair # count gets changed to pair id
                                    elif type(record[field]) is tuple:
                                        columns[field][ph_cnt] = record[field][pair]
                                    else:
                                        columns[field][ph_cnt] = record[field]
                            # Add per Photon Fields
                            for field in photon.keys():
                                if field in columns:
                                    columns[field][ph_cnt] = photon[field]
                            # Add Ancillary Extent Fields
                            for field in extent_field_dictionary:
                                columns[field][ph_cnt] = extent_field_dictionary[field][pair]
                            # Add Ancillary Extent Fields
                            for field in photon_field_dictionary:
                                columns[field][ph_cnt] = photon_field_dictionary[field][ph_index]
                            # Goto Next Photon
                            ph_cnt += 1
                            ph_index += 1
                    # Rename Count Column to Pair Column
                    columns["pair"] = columns.pop("count")

                    # Delete Extent ID Column
                    if "extent_id" in columns:
                        del columns["extent_id"]

                    # Capture Time to Flatten
                    profiles["flatten"] = time.perf_counter() - tstart_flatten

                    # Create DataFrame
                    gdf = __todataframe(columns, lat_key="latitude", lon_key="longitude")

                    # Calculate Spot Column
                    gdf['spot'] = gdf.apply(lambda row: __calcspot(row["sc_orient"], row["track"], row["pair"]), axis=1)

                    # Return Response
                    profiles[atl03sp.__name__] = time.perf_counter() - tstart
                    return gdf
                else:
                    logger.debug("No photons returned")
            else:
                logger.debug("No response returned")

    # Handle Runtime Errors
    except RuntimeError as e:
        logger.critical(e)

    # Error or No Data
    return __emptyframe()

#
#  H5
#
def h5 (dataset, resource, asset=DEFAULT_ASSET, datatype=sliderule.datatypes["DYNAMIC"], col=0, startrow=0, numrows=ALL_ROWS):
    '''
    Reads a dataset from an HDF5 file and returns the values of the dataset in a list

    This function provides an easy way for locally run scripts to get direct access to HDF5 data stored in a cloud environment.
    But it should be noted that this method is not the most efficient way to access remote H5 data, as the data is accessed one dataset at a time.
    The ``h5p`` api is the preferred solution for reading multiple datasets.

    One of the difficulties in reading HDF5 data directly from a Python script is converting the format of the data as it is stored in HDF5 to a data
    format that is easy to use in Python.  The compromise that this function takes is that it allows the user to supply the desired data type of the
    returned data via the **datatype** parameter, and the function will then return a **numpy** array of values with that data type.

    The data type is supplied as a ``sliderule.datatypes`` enumeration:

    - ``sliderule.datatypes["TEXT"]``: return the data as a string of unconverted bytes
    - ``sliderule.datatypes["INTEGER"]``: return the data as an array of integers
    - ``sliderule.datatypes["REAL"]``: return the data as an array of double precision floating point numbers
    - ``sliderule.datatypes["DYNAMIC"]``: return the data in the numpy data type that is the closest match to the data as it is stored in the HDF5 file

    Parameters
    ----------
        dataset:    str
                    full path to dataset variable (e.g. ``/gt1r/geolocation/segment_ph_cnt``)
        resource:   str
                    HDF5 filename
        asset:      str
                    data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)
        datatype:   int
                    the type of data the returned dataset list should be in (datasets that are naturally of a different type undergo a best effort conversion to the specified data type before being returned)
        col:        int
                    the column to read from the dataset for a multi-dimensional dataset; if there are more than two dimensions, all remaining dimensions are flattened out when returned.
        startrow:   int
                    the first row to start reading from in a multi-dimensional dataset (or starting element if there is only one dimension)
        numrows:    int
                    the number of rows to read when reading from a multi-dimensional dataset (or number of elements if there is only one dimension); if **ALL_ROWS** selected, it will read from the **startrow** to the end of the dataset.

    Returns
    -------
    numpy array
        dataset values

    Examples
    --------
        >>> segments    = icesat2.h5("/gt1r/land_ice_segments/segment_id",  resource, asset)
        >>> heights     = icesat2.h5("/gt1r/land_ice_segments/h_li",        resource, asset)
        >>> latitudes   = icesat2.h5("/gt1r/land_ice_segments/latitude",    resource, asset)
        >>> longitudes  = icesat2.h5("/gt1r/land_ice_segments/longitude",   resource, asset)
        >>> df = pd.DataFrame(data=list(zip(heights, latitudes, longitudes)), index=segments, columns=["h_mean", "latitude", "longitude"])
    '''
    tstart = time.perf_counter()
    datasets = [ { "dataset": dataset, "datatype": datatype, "col": col, "startrow": startrow, "numrows": numrows } ]
    values = h5p(datasets, resource, asset=asset)
    if len(values) > 0:
        profiles[h5.__name__] = time.perf_counter() - tstart
        return values[dataset]
    else:
        return numpy.empty(0)

#
#  Parallel H5
#
def h5p (datasets, resource, asset=DEFAULT_ASSET):
    '''
    Reads a list of datasets from an HDF5 file and returns the values of the dataset in a dictionary of lists.

    This function is considerably faster than the ``icesat2.h5`` function in that it not only reads the datasets in
    parallel on the server side, but also shares a file context between the reads so that portions of the file that
    need to be read multiple times do not result in multiple requests to S3.

    For a full discussion of the data type conversion options, see `h5 <../user_guide/ICESat-2.html#h5>`_.

    Parameters
    ----------
        datasets:   dict
                    list of full paths to dataset variable (e.g. ``/gt1r/geolocation/segment_ph_cnt``); see below for additional parameters that can be added to each dataset
        resource:   str
                    HDF5 filename
        asset:      str
                    data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)

    Returns
    -------
    dict
        numpy arrays of dataset values, where the keys are the dataset names

        The `datasets` dictionary can optionally contain the following elements per entry:

        * "valtype" (int): the type of data the returned dataset list should be in (datasets that are naturally of a different type undergo a best effort conversion to the specified data type before being returned)
        * "col" (int): the column to read from the dataset for a multi-dimensional dataset; if there are more than two dimensions, all remaining dimensions are flattened out when returned.
        * "startrow" (int): the first row to start reading from in a multi-dimensional dataset (or starting element if there is only one dimension)
        * "numrows" (int): the number of rows to read when reading from a multi-dimensional dataset (or number of elements if there is only one dimension); if **ALL_ROWS** selected, it will read from the **startrow** to the end of the dataset.

    Examples
    --------
        >>> from sliderule import icesat2
        >>> icesat2.init(["127.0.0.1"], False)
        >>> datasets = [
        ...         {"dataset": "/gt1l/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt1r/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt2l/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt2r/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt3l/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt3r/land_ice_segments/h_li", "numrows": 5}
        ...     ]
        >>> rsps = icesat2.h5p(datasets, "ATL06_20181019065445_03150111_003_01.h5", "atlas-local")
        >>> print(rsps)
        {'/gt2r/land_ice_segments/h_li': array([45.3146427 , 45.27640582, 45.23608027, 45.21131015, 45.15692304]),
         '/gt2l/land_ice_segments/h_li': array([45.35118977, 45.33535027, 45.27195617, 45.21816889, 45.18534204]),
         '/gt1l/land_ice_segments/h_li': array([45.68811156, 45.71368944, 45.74234326, 45.74614113, 45.79866465]),
         '/gt3l/land_ice_segments/h_li': array([45.29602321, 45.34764226, 45.31430979, 45.31471701, 45.30034622]),
         '/gt1r/land_ice_segments/h_li': array([45.72632446, 45.76512574, 45.76337375, 45.77102473, 45.81307948]),
         '/gt3r/land_ice_segments/h_li': array([45.14954134, 45.18970635, 45.16637644, 45.15235916, 45.17135806])}
    '''
    # Latch Start Time
    tstart = time.perf_counter()

    # Baseline Request
    rqst = {
        "asset" : asset,
        "resource": resource,
        "datasets": datasets,
    }

    # Read H5 File
    try:
        rsps = sliderule.source("h5p", rqst, stream=True)
    except RuntimeError as e:
        logger.critical(e)
        rsps = []

    # Build Record Data
    results = {}
    for result in rsps:
        results[result["dataset"]] = __get_values(result["data"], result["datatype"], result["size"])

    # Update Profiles
    profiles[h5p.__name__] = time.perf_counter() - tstart

    # Return Results
    return results

#
# Format Region Specification
#
def toregion(source, tolerance=0.0, cellsize=0.01, n_clusters=1):
    '''
    Convert a GeoJSON representation of a set of geospatial regions into a list of lat,lon coordinates and raster image recognized by SlideRule

    Parameters
    ----------
        filename:   str
                    file name of GeoJSON formatted regions of interest, file **must** have name with the .geojson suffix
                    file name of ESRI Shapefile formatted regions of interest, file **must** have name with the .shp suffix
        tolerance:  float
                    tolerance used to simplify complex shapes so that the number of points is less than the limit (a tolerance of 0.001 typically works for most complex shapes)
        cellsize:   float
                    size of pixel in degrees used to create the raster image of the polygon
        clusters:   int
                    number of clusters of polygons to create when breaking up the request to CMR

    Returns
    -------
    dict
        a list of longitudes and latitudes containing the region of interest that can be used for the **poly** and **raster** parameters in a processing request to SlideRule.

        region = {"poly": [{"lat": <lat1>, "lon": <lon1>, ... }], "clusters": [{"lat": <lat1>, "lon": <lon1>, ... }, {"lat": <lat1>, "lon": <lon1>, ... }, ...], "raster": {"data": <geojson file as string>, "length": <length of geojson file>, "cellsize": <parameter cellsize>}}

    Examples
    --------
        >>> from sliderule import icesat2
        >>> # Region of Interest #
        >>> region_filename = sys.argv[1]
        >>> region = icesat2.toregion(region_filename)
        >>> # Configure SlideRule #
        >>> icesat2.init("slideruleearth.io", False)
        >>> # Build ATL06 Request #
        >>> parms = {
        ...     "poly": region["poly"],
        ...     "srt": icesat2.SRT_LAND,
        ...     "cnf": icesat2.CNF_SURFACE_HIGH,
        ...     "ats": 10.0,
        ...     "cnt": 10,
        ...     "len": 40.0,
        ...     "res": 20.0,
        ...     "maxi": 1
        ... }
        >>> # Get ATL06 Elevations
        >>> atl06 = icesat2.atl06p(parms)
    '''

    tstart = time.perf_counter()
    tempfile = "temp.geojson"

    if isinstance(source, geopandas.GeoDataFrame):
        # user provided GeoDataFrame instead of a file
        gdf = source
        # Convert to geojson file
        gdf.to_file(tempfile, driver="GeoJSON")
        with open(tempfile, mode='rt') as file:
            datafile = file.read()
        os.remove(tempfile)

    elif isinstance(source, list) and (len(source) >= 4) and (len(source) % 2 == 0):
        # create lat/lon lists
        if len(source) == 4: # bounding box
            lons = [source[0], source[2], source[2], source[0], source[0]]
            lats = [source[1], source[1], source[3], source[3], source[1]]
        elif len(source) > 4: # polygon list
            lons = [source[i] for i in range(1,len(source),2)]
            lats = [source[i] for i in range(0,len(source),2)]

        # create geodataframe
        p = Polygon([point for point in zip(lons, lats)])
        gdf = geopandas.GeoDataFrame(geometry=[p], crs=EPSG_MERCATOR)

        # Convert to geojson file
        gdf.to_file(tempfile, driver="GeoJSON")
        with open(tempfile, mode='rt') as file:
            datafile = file.read()
        os.remove(tempfile)

    elif isinstance(source, str) and (source.find(".shp") > 1):
        # create geodataframe
        gdf = geopandas.read_file(source)
        # Convert to geojson file
        gdf.to_file(tempfile, driver="GeoJSON")
        with open(tempfile, mode='rt') as file:
            datafile = file.read()
        os.remove(tempfile)

    elif isinstance(source, str) and (source.find(".geojson") > 1):
        # create geodataframe
        gdf = geopandas.read_file(source)
        with open(source, mode='rt') as file:
            datafile = file.read()

    else:
        raise TypeError("incorrect filetype: please use a .geojson, .shp, or a geodataframe")


    # If user provided raster we don't have gdf, geopandas cannot easily convert it
    polygon = clusters = None
    if gdf is not None:
        # simplify polygon
        if(tolerance > 0.0):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                gdf = gdf.buffer(tolerance)
                gdf = gdf.simplify(tolerance)

        # generate polygon
        polygon = __gdf2poly(gdf)

        # generate clusters
        clusters = []
        if n_clusters > 1:
            # pull out centroids of each geometry object
            if "CenLon" in gdf and "CenLat" in gdf:
                X = numpy.column_stack((gdf["CenLon"], gdf["CenLat"]))
            else:
                s = gdf.centroid
                X = numpy.column_stack((s.x, s.y))
            # run k means clustering algorithm against polygons in gdf
            kmeans = KMeans(n_clusters=n_clusters, init='k-means++', random_state=5, max_iter=400)
            y_kmeans = kmeans.fit_predict(X)
            k = geopandas.pd.DataFrame(y_kmeans, columns=['cluster'])
            gdf = gdf.join(k)
            # build polygon for each cluster
            for n in range(n_clusters):
                c_gdf = gdf[gdf["cluster"] == n]
                c_poly = __gdf2poly(c_gdf)
                clusters.append(c_poly)

    # update timing profiles
    profiles[toregion.__name__] = time.perf_counter() - tstart

    # return region
    return {
        "gdf": gdf,
        "poly": polygon, # convex hull of polygons
        "clusters": clusters, # list of polygon clusters for cmr request
        "raster": {
            "data": datafile, # geojson file
            "length": len(datafile), # geojson file length
            "cellsize": cellsize  # untis are in crs/projection
        }
    }

#
# Get Version
#
def get_version ():
    '''
    Get the version information for the running servers and Python client

    Returns
    -------
    dict
        dictionary of version information
    '''
    return sliderule.get_version()



#
# Build index file list and create vrt
import geopandas as gpd
import numpy as np
import pandas as pd
import pystac


###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # catalog_stac = 'https://pgc-opendata-dems.s3.us-west-2.amazonaws.com/pgc-data-stac.json'
    # cat = pystac.read_file(catalog_stac)
    # arcticdem_collection = cat.get_child('arcticdem')
    # mosaic_collection = arcticdem_collection.get_child('arcticdem-mosaics')

    collection_stac = 'https://pgc-opendata-dems.s3.us-west-2.amazonaws.com/arcticdem/mosaics/v3.0/2m.json'
    col = pystac.read_file(collection_stac)

    item_list = []
    cnt = 0

    for link in col.links:
        if link.rel == pystac.RelType.CHILD:
            subcat = pystac.read_file(link.target)
            print(subcat)
            # if cnt == 2:
            #     break
            # cnt += 1

            for _link in subcat.links:
                if _link.rel == pystac.RelType.CHILD:
                    item = pystac.read_file(_link.target)
                    dem = item.to_dict()['assets']['dem']['href']
                    dem = dem.replace(".", "", 1)
                    path = link.target.replace("https:/", "/vsis3")
                    path = path.replace(".s3.us-west-2.amazonaws.com", "", 1)
                    path = path.replace(".json", dem)
                    item_list.append(path)

    print(f"Number of features: {len(item_list)}")

    with open('/data/ArcticDem/mosaic_vrt_list.txt', 'w') as f:
        for line in item_list:
            f.write(f"{line}\n")




#
# Uses the icesat2.h5p api to read a dataset from an H5 file and write the contents to a file
#

import sys
import logging
from sliderule import icesat2
from utils import parse_command_line

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # set script defaults
    cfg = {
        "url":          'localhost',
        "organization": None,
        "asset":        'atlas-local',
        "dataset":      '/gt2l/heights/h_ph',
        "resource":     'ATL03_20181017222812_02950102_005_01.h5',
        "col":          0,
        "startrow":     0,
        "numrows":      -1
    }

    # parse configuration parameters
    parse_command_line(sys.argv, cfg)

    # Configure Logging #
    logging.basicConfig(level=logging.INFO)

    # Configure SlideRule #
    icesat2.init(cfg["url"], True, organization=cfg["organization"])

    # Read Dataset #
    datasets = [ {"dataset": cfg["dataset"], "col": cfg["col"], "startrow": cfg["startrow"], "numrows": cfg["numrows"]} ]
    rawdata = icesat2.h5p(datasets, cfg["resource"], cfg["asset"])
    print(rawdata)

    # Write Dataset to File #
    filename = cfg["dataset"][cfg["dataset"].rfind("/")+1:]
    f = open(filename + ".bin", 'w+b')
    f.write(bytearray(rawdata[cfg["dataset"]]))
    f.close()



#
# Requests SlideRule to process the provided region of interest twice:
# (1) perform ATL06-SR algorithm to calculate elevations in region
# (2) retrieve photon counts from each ATL06 extent
#
# The region of interest is passed in as a json file that describes the
# geospatial polygon for the region. An example json object is below:
#
# {
#   "region": [{"lon": 86.269733430535638, "lat": 28.015965655545852},
#              {"lon": 86.261403224371804, "lat": 27.938668666352985},
#              {"lon": 86.302412923741514, "lat": 27.849318271202186},
#              {"lon": 86.269733430535638, "lat": 28.015965655545852}]
# }
#

import sys
import logging
import geopandas as gpd
import matplotlib.pyplot as plt
from sliderule import icesat2
from utils import initialize_client, display_statistics

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # Configure Logging
    logging.basicConfig(level=logging.INFO)

    # Initialize Client #
    parms, cfg = initialize_client(sys.argv)

    # Get ATL06 Elevations
    atl06 = icesat2.atl06p(parms, cfg["asset"])

    # Display Statistics
    display_statistics(atl06, "elevations")

    # Exit Early If No Photos
    if len(atl06) <= 0:
        sys.exit(0)

    # Calculate Extent
    lons = [p["lon"] for p in parms["poly"]]
    lats = [p["lat"] for p in parms["poly"]]
    lon_margin = (max(lons) - min(lons)) * 0.1
    lat_margin = (max(lats) - min(lats)) * 0.1
    extent = (min(lons) - lon_margin, max(lons) + lon_margin, min(lats) - lat_margin, max(lats) + lat_margin)

    # Create Plot
    fig = plt.figure(num=None, figsize=(24, 12))

    # Plot ATL06 Ground Tracks
    ax1 = plt.subplot(231)
    ax1.set_title("Zoomed ATL06 Ground Tracks")
    atl06.plot(ax=ax1, column='h_mean', cmap='plasma', markersize=0.5)
    ax1.plot(lons, lats, linewidth=1.5, color='r', zorder=2)

    # Plot ATL06 Along Track Slope
    ax2 = plt.subplot(232)
    ax2.set_title("Zoomed ATL06 Along Track Slope")
    atl06.plot(ax=ax2, column='dh_fit_dx', cmap='inferno', markersize=0.5)
    ax2.plot(lons, lats, linewidth=1.5, color='r', zorder=2)

    # Plot Global View
    ax3 = plt.subplot(233)
    ax3.set_title("Global Reference")
    world = gpd.read_file(gpd.datasets.get_path('naturalearth_lowres'))
    world.plot(ax=ax3, color='0.8', edgecolor='black')
    atl06.plot(ax=ax3, marker='o', color='red', markersize=2.5, zorder=3)
    ax3.set_xlim(-180,180)
    ax3.set_ylim(-90,90)
    ax3.set_aspect('equal', adjustable='box')

    # Plot Number of Fit Photons per ATL06 Ground Tracks
    ax4 = plt.subplot(234)
    atl06.hist("n_fit_photons", bins=100, ax=ax4)
    ax4.set_title("Number of Fit Photons")

    # Plot Final Window Size per ATL06 Ground Tracks
    ax5 = plt.subplot(235)
    atl06.hist("w_surface_window_final", bins=100, ax=ax5)
    ax5.set_title("Final Window Size")

    # Plot Number of Fit Photons per ATL06 Ground Tracks
    ax6 = plt.subplot(236)
    atl06.hist("rms_misfit", bins=100, ax=ax6)
    ax6.set_title("RMS of Fit")

    # Show Plot
    plt.show()


import sys
import json
import logging
import sliderule
from sliderule import icesat2
from utils import initialize_client, display_statistics

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # Configure Logging
    logging.basicConfig(level=logging.INFO)

    # Initialize Client
    parms, cfg = initialize_client(sys.argv)

    # Query Version
    rsps = sliderule.source("version", {})

    # Display Version Information
    print(json.dumps(rsps, indent=4))


#
#   Uses the "event" endpoint to capture a set of traces
#   and produce human readable results
#

import sys
import pandas
import sliderule
from utils import parse_command_line

###############################################################################
# GLOBALS
###############################################################################

TRACE_ORIGIN = 0
TRACE_START = 1
TRACE_STOP = 2
LOG = 1
TRACE = 2
METRIC = 4
COLOR_MAP = [8421631, 8454143, 8454016, 16777088, 16744703, 16777215]

names = {} # dictionary of unique trace names
traces = {} # dictionary of unique traces
origins = [] # list of highest level "root" traces

###############################################################################
# FUNCTIONS
###############################################################################

def display_trace(trace, depth):
    # Correct missing stops
    if trace["stop"] == None:
        trace["stop"] = trace["start"]
    # Get values of trace
    trace_id        = trace["start"]['id']
    thread_id       = trace["start"]['tid']
    start_time      = trace["start"]['time']
    stop_time       = trace["stop"]['time']
    sec_from_origin = start_time / 1e3
    sec_duration    = (stop_time - start_time) / 1e3
    dt              = sliderule.gps2utc(sec_from_origin)
    name            = trace["start"]['name']
    attributes      = trace["start"]['attr']
    # Print trace
    print('{} ({:7.3f} sec):{:{indent}}{:{width}} <{}> {} [{}]'.format(dt, sec_duration, "", str(name), thread_id, attributes, trace_id, indent=depth, width=30-depth))
    # Recurse on children
    for child in trace["children"]:
        display_trace(child, depth + 2)

def write_sta_events(filename, df):
    f = open(filename, "w")
    for index,row in df.iterrows():
        f.write("%08d %08X %.3f ms\n" % (index, int(row["id"] + (int(row["edge"]) << 14)), row["delta"]))
    f.close()

def write_sta_setup(filename, perf_ids):
    f = open(filename, "w")
    f.write("[PerfID Table]\n")
    f.write("Version=1\n")
    f.write("Auto Hide=True\n")
    f.write("Auto Set Bit For Exit=True\n")
    f.write("Bit To Set For Exit=14\n")
    f.write("Number of PerfIDs=%d\n" % (len(perf_ids)))
    index = 0
    for name in perf_ids:
        perf_id = int(perf_ids[name]["id"])
        depth = int(perf_ids[name]["depth"])
        if(depth > len(COLOR_MAP)):
            depth = len(COLOR_MAP)
        f.write("[PerfID %d]\n" % (index))
        f.write("Name=%s\n" % (name))
        f.write("Entry ID=%08X\n" % perf_id)
        f.write("Exit ID=%08X\n" % (int(perf_id + (int(1) << 14))))
        f.write("Calc CPU=True\n")
        f.write("Color=%d\n" % (COLOR_MAP[depth - 1]))
        f.write("Hide=False\n")
        f.write("DuplicateEdgeWarningsDisabled=False\n")
        index += 1
    f.close()

def build_event_list(trace, depth, max_depth, names, events, perf_ids):
    # Get Perf ID
    name = trace["name"]
    perf_id = names.index(name)
    perf_ids[name] = {"id": perf_id, "depth": depth}
    # Append Events
    try:
        events.append({"id": perf_id, "time": trace["start"]["time"], "edge": 0})
        events.append({"id": perf_id, "time": trace["stop"]["time"], "edge": 1})
    except:
        pass
    # Recurse on Children
    if (depth < max_depth) or (max_depth == 0):
        for child in trace["children"]:
            build_event_list(child, depth + 1, max_depth, names, events, perf_ids)

def console_output(origins):
    # Output traces to console
    for trace in origins:
        display_trace(trace, 1)

def sta_output(idlist, depth, names, traces):
    global origins
    # Build list of events and names
    events = []
    perf_ids = {}
    if len(idlist) > 0:
        for trace_id in idlist:
            build_event_list(traces[trace_id], 1, depth, names, events, perf_ids)
    else:
        for trace in origins:
            build_event_list(trace, 1, depth, names, events, perf_ids)
    # Build and sort data frame
    df = pandas.DataFrame(events)
    df = df.sort_values("time")
    # Build delta times
    df["delta"] = df["time"].diff()
    df.at[0, "delta"] = 0.0
    # Write out data frame as sta events
    write_sta_events("pytrace.txt", df)
    write_sta_setup("pytrace.PerfIDSetup", perf_ids)

def process_event(rec):
    global names, traces, origins
    # Populate traces dictionary
    if rec["type"] == LOG:
        print('%s:%s:%s' % (rec["ipv4"], rec["name"], rec["attr"]))
    elif rec["type"] == TRACE:
        trace_id = rec['id']
        if rec["flags"] & TRACE_START:
            if trace_id not in traces.keys():
                # Populate start of span
                name = str(rec['name']) + "." + str(rec['tid'])
                traces[trace_id] = {"id": trace_id, "name": name, "start": rec, "stop": None, "children": []}
                # Link to parent
                parent_trace_id = rec['parent']
                if parent_trace_id in traces.keys():
                    traces[parent_trace_id]["children"].append(traces[trace_id])
                else:
                    origins.append(traces[trace_id])
                # Populate name
                names[name] = True
            else:
                print('warning: double start for %s' % (rec['name']))
        elif rec["flags"] & TRACE_STOP:
            if trace_id in traces.keys():
                # Populate stop of span
                traces[trace_id]["stop"] = rec
            else:
                print('warning: stop without start for %s' % (rec['name']))

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # Default Parameters
    parms = {
        "url": "localhost",
        "organization": None,
        "fmt": "console",
        "depth": 0,
        "ids": []
    }

    # Override Parameters
    parms = parse_command_line(sys.argv, parms)

    # Default Request
    rqst = {
        "type": LOG | TRACE,
        "level" : "INFO",
        "duration": 30
    }

    # Override Request
    rqst = parse_command_line(sys.argv, rqst)

    # Set URL and Organization
    sliderule.set_url(parms["url"])
    sliderule.authenticate(parms["organization"])

    # Connect to SlideRule
    rsps = sliderule.source("event", rqst, stream=True, callbacks={'eventrec': process_event})

    # Flatten names to get indexes
    names = list(names)

    # Run commanded operation
    if parms["fmt"] == "console":
        console_output(origins)
    elif parms["fmt"] == "sta":
        sta_output(parms["ids"], parms["depth"], names, traces)



import time
import json
from sliderule import icesat2

#
# Globals
#
tstart = 0.0

#
# Parse Command Line
#
def parse_command_line(args, cfg):

    i = 1
    for i in range(1,len(args)):
        for entry in cfg:
            if args[i] == '--'+entry:
                if type(cfg[entry]) is str or cfg[entry] == None:
                    if args[i + 1] == "None":
                        cfg[entry] = None
                    else:
                        cfg[entry] = args[i + 1]
                elif type(cfg[entry]) is list:
                    if args[i + 1] == "None":
                        cfg[entry] = None
                    else:
                        l = []
                        while (i + 1) < len(args) and '--' not in args[i + 1]:
                            if args[i + 1].isnumeric():
                                l.append(int(args[i + 1]))
                            else:
                                l.append(args[i + 1])
                            i += 1
                        cfg[entry] = l
                elif type(cfg[entry]) is int:
                    if args[i + 1] == "None":
                        cfg[entry] = None
                    else:
                        cfg[entry] = int(args[i + 1])
                elif type(cfg[entry]) is bool:
                    if args[i + 1] == "None":
                        cfg[entry] = None
                    elif args[i + 1] == "True" or args[i + 1] == "true":
                        cfg[entry] = True
                    elif args[i + 1] == "False" or args[i + 1] == "false":
                        cfg[entry] = False

#
# Initialize Client
#
def initialize_client(args):

    global tstart

    # Set Script Defaults
    cfg = {
        "domain":                   'localhost',
        "organization":             None,
        "asset":                    'atlas-local',
        "region":                   'examples/grandmesa.geojson',
        "resource":                 'ATL03_20181017222812_02950102_005_01.h5',
        "raster":                   True,
        "atl08_class":              [],
        "yapc.score":               0,
        "yapc.knn":                 0,
        "yapc.min_knn":             5,
        "yapc.win_h":               6.0,
        "yapc.win_x":               15.0,
        "yapc.version":             0,
        "srt":                      icesat2.SRT_LAND,
        "cnf":                      icesat2.CNF_SURFACE_HIGH,
        "ats":                      10.0,
        "cnt":                      10,
        "len":                      40.0,
        "res":                      20.0,
        "maxi":                     1,
        "atl03_geo_fields":         [],
        "atl03_ph_fields":          [],
        "profile":                  True,
        "verbose":                  True,
        "timeout":                  0,
        "rqst-timeout":             0,
        "node-timeout":             0,
        "read-timeout":             0,
        "output.path":              None,
        "output.format":            "native",
        "output.open_on_complete":  False
    }

    # Parse Configuration Parameters
    parse_command_line(args, cfg)

    # Configure SlideRule
    icesat2.init(cfg["domain"], cfg["verbose"], organization=cfg["organization"])

    # Build Initial Parameters
    parms = {
        "srt":  cfg['srt'],
        "cnf":  cfg['cnf'],
        "ats":  cfg['ats'],
        "cnt":  cfg['cnt'],
        "len":  cfg['len'],
        "res":  cfg['res'],
        "maxi": cfg['maxi'],
    }

    # Region of Interest
    if cfg["region"]:
        region = icesat2.toregion(cfg["region"])
        parms["poly"] = region['poly']
        if cfg["raster"]:
            parms["raster"] = region['raster']

    # Add Ancillary Fields
    if len(cfg['atl03_geo_fields']) > 0:
        parms['atl03_geo_fields'] = cfg['atl03_geo_fields']
    if len(cfg['atl03_ph_fields']) > 0:
        parms['atl03_ph_fields'] = cfg['atl03_ph_fields']

    # Add ATL08 Classification
    if len(cfg['atl08_class']) > 0:
        parms['atl08_class'] = cfg['atl08_class']

    # Add YAPC Parameters
    if cfg["yapc.version"] > 0:
        parms["yapc"] = { "score": cfg["yapc.score"],
                          "knn": cfg["yapc.knn"],
                          "min_knn": cfg["yapc.min_knn"],
                          "win_h": cfg["yapc.win_h"],
                          "win_x": cfg["yapc.win_x"],
                          "version": cfg["yapc.version"] }

    # Provide Timeouts
    if cfg["timeout"] > 0:
        parms["timeout"] = cfg["timeout"]
        parms["rqst-timeout"] = cfg["timeout"]
        parms["node-timeout"] = cfg["timeout"]
        parms["read-timeout"] = cfg["timeout"]
    if cfg["rqst-timeout"] > 0:
        parms["rqst-timeout"] = cfg["rqst-timeout"]
    if cfg["node-timeout"] > 0:
        parms["node-timeout"] = cfg["node-timeout"]
    if cfg["read-timeout"] > 0:
        parms["read-timeout"] = cfg["read-timeout"]

    # Add Output Options
    if cfg["output.path"]:
        parms["output"] = { "path": cfg["output.path"],
                            "format": cfg["output.format"],
                            "open_on_complete": cfg["output.open_on_complete"] }
    # Latch Start Time
    tstart = time.perf_counter()

    # Return Parameters and Configuration
    return parms, cfg

#
# Display Statistics
#
def display_statistics(gdf, name):

    global tstart

    perf_duration = time.perf_counter() - tstart
    print("Completed in {:.3f} seconds of wall-clock time".format(perf_duration))
    if len(gdf) > 0:
        print("Reference Ground Tracks: {}".format(gdf["rgt"].unique()))
        print("Cycles: {}".format(gdf["cycle"].unique()))
        print("Received {} {}".format(len(gdf), name))
    else:
        print("No {} were returned".format(name))

    print("\nTiming Profiles")
    for key in icesat2.profiles:
        print("{:16}: {:.6f} secs".format(key, icesat2.profiles[key]))

#
# Pretty Print JSON
#
def pprint(obj):
    print(json.dumps(obj, indent=2))



import sys
from datetime import date
from sliderule import ipxapi
from sliderule import icesat2
from utils import parse_command_line
import matplotlib.pyplot as plt
import icepyx

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    today = date.today()

    # set script defaults
    scfg = {
        "url": 'localhost',
        "organization": None,
        "asset": 'atlas-local'
    }

    # set icepx defaults
    icfg = {
        "short_name": 'ATL03',
        "spatial_extent": 'tests/data/grandmesa.shp',
        "date_range": ['2018-01-01', "{}-{}-{}".format(today.year, today.month, today.day)],
        "cycles": None,
        "tracks": None
    }

    # set processing parameter defaults
    parms = {
        "srt": icesat2.SRT_LAND,
        "cnf": icesat2.CNF_SURFACE_HIGH,
        "ats": 10.0,
        "cnt": 10,
        "len": 40.0,
        "res": 20.0,
        "maxi": 1
    }

    # get command line parameters
    parse_command_line(sys.argv, icfg)
    parse_command_line(sys.argv, scfg)
    parse_command_line(sys.argv, parms)

    # create icepx region
    iregion = icepyx.Query(icfg["short_name"], icfg["spatial_extent"], icfg["date_range"], cycles=icfg["cycles"], tracks=icfg["tracks"])

    # visualize icepx region
    # iregion.visualize_spatial_extent()

    # display summary information
    iregion.product_summary_info()
    # print("Available Granules:", iregion.avail_granules())
    # print("Available Granule IDs:", iregion.avail_granules(ids=True))

    # initialize sliderule api
    icesat2.init(scfg["url"], verbose=True, organization=scfg["organization"])

    # generate sliderule atl06 elevations
    # parms["poly"] = icesat2.toregion(icfg["spatial_extent"])["poly"]
    atl06_sr = ipxapi.atl06p(iregion, parms, scfg["asset"])

    # create plot
    f, ax = plt.subplots()
    vmin, vmax = atl06_sr['h_mean'].quantile((0.02, 0.98))
    atl06_sr.plot(ax=ax, column='h_mean', cmap='inferno', s=0.1, vmin=vmin, vmax=vmax)
    plt.show()



#
# Connects to SlideRule server node at provided url and prints the last
# 1K log messages to local terminal.
#
# This bypasses service discovery and goes directly to the server node.
#
# Use query_services.py to get list of server node IP addresses
#
import sys
import logging
from sliderule import sliderule
from sliderule import icesat2
from utils import parse_command_line

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # Set Script Defaults
    cfg = {
        "url":          'localhost',
        "organization": None,
        "monitor":      'EventMonitor'
    }

    # Parse Configuration Parameters
    parse_command_line(sys.argv, cfg)

    # configure logging
    logging.basicConfig(level=logging.INFO)

    # Initialize ICESat2/SlideRule Package
    icesat2.init(cfg["url"], True, organization=cfg["organization"])

    # Build Logging Request
    rqst = {
        "monitor": cfg["monitor"]
    }

    # Retrieve logs
    rsps = sliderule.source("tail", rqst, stream=False)

    # Display logs
    for rsp in rsps:
        print(rsp, end='')



#
# Connects to SlideRule server node at provided url and prints the metrics
# associated with the queried attribute.
#
# This bypasses service discovery and goes directly to the server node.
#
# Use query_services.py to get list of server node IP addresses
#
import sys
import logging
import json
from sliderule import sliderule
from sliderule import icesat2

###############################################################################
# GLOBAL CODE
###############################################################################

# configure logging
logging.basicConfig(level=logging.INFO)

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    url = "127.0.0.1"
    attr = "SourceEndpoint"
    organization = None

    # Override server URL from command line
    if len(sys.argv) > 1:
        url = sys.argv[1]

    # Override organization
    if len(sys.argv) > 2:
        organization = sys.argv[2]

    # Override group of endpoints to query
    if len(sys.argv) > 3:
        attr = sys.argv[3]

    # Initialize ICESat2/SlideRule package
    icesat2.init(url, True, organization=organization)

    # Retrieve metrics
    rsps = sliderule.source("metric", { "attr": attr }, stream=False)

    # Display metrics
    print(json.dumps(rsps, indent=2))



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from sliderule import icesat2
import logging

###############################################################################
# GLOBALS
###############################################################################

# create logger
logger = logging.getLogger(__name__)

###############################################################################
# APIs
###############################################################################

#
#  ICEPYX ATL06
#
def atl06p(ipx_region, parm, asset=icesat2.DEFAULT_ASSET):
    """
    Performs ATL06-SR processing in parallel on ATL03 data and returns gridded elevations.  The list of granules to be processed is identified by the ipx_region object.

    See the `atl06p <../api_reference/icesat2.html#atl06p>`_ function for more details.

    Parameters
    ----------
        ipx_region: Query
                    icepyx region object defining the query of granules to be processed
        parm:       dict
                    parameters used to configure ATL06-SR algorithm processing (see `Parameters <../user_guide/ICESat-2.html#parameters>`_)
        asset:      str
                    data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)

    Returns
    -------
    GeoDataFrame
        gridded elevations (see `Elevations <../user_guide/ICESat-2.html#elevations>`_)
    """
    try:
        version = ipx_region.product_version
        resources = ipx_region.avail_granules(ids=True)[0]
    except:
        logger.critical("must supply an icepyx query as region")
        return icesat2.__emptyframe()
    # try to get the subsetting region
    if ipx_region.extent_type in ('bbox','polygon'):
        parm.update({'poly': to_region(ipx_region)})

    return icesat2.atl06p(parm, asset, version=version, resources=resources)

#
#  ICEPYX ATL03
#
def atl03sp(ipx_region, parm, asset=icesat2.DEFAULT_ASSET):
    """
    Performs ATL03 subsetting in parallel on ATL03 data and returns photon segment data.

    See the `atl03sp <../api_reference/icesat2.html#atl03sp>`_ function for more details.

    Parameters
    ----------
        ipx_region: Query
                    icepyx region object defining the query of granules to be processed
        parms:      dict
                    parameters used to configure ATL03 subsetting (see `Parameters <../user_guide/ICESat-2.html#parameters>`_)
        asset:      str
                    data source asset (see `Assets <../user_guide/ICESat-2.html#assets>`_)

    Returns
    -------
    list
        ATL03 segments (see `Photon Segments <../user_guide/ICESat-2.html#photon-segments>`_)
    """
    try:
        version = ipx_region.product_version
        resources = ipx_region.avail_granules(ids=True)[0]
    except:
        logger.critical("must supply an icepyx query as region")
        return icesat2.__emptyframe()
    # try to get the subsetting region
    if ipx_region.extent_type in ('bbox','polygon'):
        parm.update({'poly': to_region(ipx_region)})

    return icesat2.atl03sp(parm, asset, version=version, resources=resources)

def to_region(ipx_region):
    """
    Extract subsetting extents from an icepyx region

    Parameters
    ----------
        ipx_region: Query
                    icepyx region object defining the query of granules to be processed

    Returns
    -------
    list
        polygon definining region of interest (can be passed into `icesat2` api functions)

    """
    if (ipx_region.extent_type == 'bbox'):
        bbox = ipx_region.spatial_extent[1]
        poly = [dict(lon=bbox[0], lat=bbox[1]),
                dict(lon=bbox[2], lat=bbox[1]),
                dict(lon=bbox[2], lat=bbox[3]),
                dict(lon=bbox[0], lat=bbox[3]),
                dict(lon=bbox[0], lat=bbox[1])]
    elif (ipx_region.extent_type == 'polygon'):
        poly = [dict(lon=ln,lat=lt) for ln,lt in zip(*ipx_region.spatial_extent[1])]
    return poly



#
# Connects to SlideRule server node at provided url and prints log messages
# as they are generated on server to local terminal.
#
# This bypasses service discovery and goes directly to the server node.
#
# Use query_services.py to get list of server node IP addresses
#

import sys
import logging
from sliderule import sliderule
from sliderule import icesat2
from utils import parse_command_line

###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':

    # Set Script Defaults
    cfg = {
        "url":          'localhost',
        "organization": None,
        "duration":     30, # seconds
        "event_type":   'LOG',
        "event_level":  'INFO'
    }

    # Parse Configuration Parameters
    parse_command_line(sys.argv, cfg)

    # configure logging
    logging.basicConfig(level=logging.INFO)

    # Initialize ICESat2/SlideRule Package
    icesat2.init(cfg["url"], True, organization=cfg["organization"])

    # Build Logging Request
    rqst = {
        "type":     cfg["event_type"],
        "level" :   cfg["event_level"],
        "duration": cfg["duration"]
    }

    # Retrieve logs
    rsps = sliderule.source("event", rqst, stream=True)



import os
from setuptools import setup, find_packages

# get long_description from README.md
with open("README.md", "r") as fh:
    long_description = fh.read()

# get install requirements
with open('requirements.txt') as fh:
    install_requires = fh.read().splitlines()

# get version
with open('version.txt') as fh:
    version = fh.read().strip()
    if version[0] == 'v':
        version = version[1:]

# list of all utility scripts to be included with package
scripts=[os.path.join('utils',f) for f in os.listdir('utils') if f.endswith('.py')]

setup(
    name='sliderule',
    author='SlideRule Developers',
    description='Python client for interacting with sliderule server',
    long_description_content_type="text/markdown",
    url='https://github.com/ICESat2-SlideRule/sliderule-python/',
    license='BSD 3-Clause',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    packages=find_packages(),
    version=version,
    install_requires=install_requires,
    scripts=scripts,
)



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import sliderule

###############################################################################
# GLOBALS
###############################################################################

# create logger
logger = logging.getLogger(__name__)

# default asset
DEFAULT_ASSET="arcticdem-s3"

###############################################################################
# APIs
###############################################################################

#
#  Get Elevation
#
def elevation (coordinate, asset=DEFAULT_ASSET):
    '''
    Get elevation from ArcticDEM at provided coordinate

    Parameters
    ----------
        coordinate :    list
                        [<longitude>, <latitude>]
        asset:          str
                        data source asset (see `Assets <../user_guide/ArcticDEM.html#assets>`_)

    Examples
    --------
        >>> from sliderule import arcticdem
        >>> arcticdem.elevation([23.14333, 70.3211])
    '''
    return elevations([coordinate])

#
#  Get Elevations
#
def elevations (coordinates, asset=DEFAULT_ASSET):
    '''
    Get elevations from ArcticDEM at provided coordinates

    Parameters
    ----------
        coordinates :   list
                        [[<longitude>, <latitude>], [<longitude>, <latitude>], ... ]
        asset:          str
                        data source asset (see `Assets <../user_guide/ArcticDEM.html#assets>`_)

    Examples
    --------
        >>> from sliderule import arcticdem
        >>> arcticdem.elevations([[164.134, 73.9291], [23.14333, 70.3211]])
    '''
    # Build Request
    rqst = {
        "dem-asset" : asset,
        "coordinates": coordinates
    }

    # Make API Processing Request
    rsps = sliderule.source("elevation", rqst, stream=False)

    # Return Response
    return rsps
